""" ACL tests for `core_linked_records.components.data.api`.
"""
from unittest import TestCase
from unittest.mock import patch, MagicMock

from core_linked_records_app.components.data import (
    api as pid_data_api,
    access_control as pid_data_acl,
)
from core_linked_records_app.components.pid_path import api as pid_path_api
from core_main_app.access_control import api as main_acl
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.components.data import api as main_data_api
from core_main_app.utils.tests_tools.MockUser import create_mock_user


class TestGetDataByPid(TestCase):
    """ACL tests for `get_data_by_pid` function."""

    def setUp(self) -> None:
        """setUp"""
        self.mock_request = MagicMock()
        self.mock_query_result = MagicMock()

    def setup_mocks(
        self, mock_get_all, mock_execute_json_query, user, owner=None
    ) -> None:
        """setup_mocks"""
        self.mock_request.user = user
        self.mock_query_result.user_id = owner.id if owner else user.id

        mock_get_all.return_value = []
        mock_execute_json_query.return_value = [self.mock_query_result]

    @patch.object(main_data_api, "execute_json_query")
    @patch.object(pid_path_api, "get_all")
    def test_superuser_can_access(self, mock_get_all, mock_execute_json_query):
        """test_superuser_can_access"""
        user = create_mock_user("1", is_superuser=True)

        self.setup_mocks(mock_get_all, mock_execute_json_query, user)

        self.assertEqual(
            pid_data_api.get_data_by_pid("mock_pid", self.mock_request),
            self.mock_query_result,
        )

    @patch.object(main_data_api, "execute_json_query")
    @patch.object(pid_path_api, "get_all")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_registered_user_not_owner_cannot_access_private(
        self,
        mock_setting,
        mock_workspace_api,  # noqa, pylint: disable=unused-argument
        mock_get_all,
        mock_execute_json_query,
    ):
        """test_registered_user_not_owner_cannot_access_private"""
        mock_setting.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = False
        user = create_mock_user("1")
        owner = create_mock_user("2")

        self.setup_mocks(mock_get_all, mock_execute_json_query, user, owner)

        with self.assertRaises(AccessControlError):
            pid_data_api.get_data_by_pid("mock_pid", self.mock_request)

    @patch.object(main_data_api, "execute_json_query")
    @patch.object(pid_path_api, "get_all")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_registered_user_not_owner_can_access_public(
        self,
        mock_settings,
        mock_workspace_api,
        mock_get_all,
        mock_execute_json_query,
    ):
        """test_registered_user_not_owner_can_access_public"""
        mock_public_workspace = MagicMock()
        self.mock_query_result.workspace = mock_public_workspace
        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = True
        mock_workspace_api.get_all_workspaces_with_read_access_by_user.return_value = [
            mock_public_workspace
        ]
        user = create_mock_user("1")
        owner = create_mock_user("2")

        self.setup_mocks(mock_get_all, mock_execute_json_query, user, owner)

        self.assertEqual(
            pid_data_api.get_data_by_pid("mock_pid", self.mock_request),
            self.mock_query_result,
        )

    @patch.object(main_data_api, "execute_json_query")
    @patch.object(pid_path_api, "get_all")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_registered_user_and_owner_can_access_private(
        self,
        mock_settings,
        mock_workspace_api,  # noqa, pylint: disable=unused-argument
        mock_get_all,
        mock_execute_json_query,
    ):
        """test_registered_user_and_owner_can_access_private"""
        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = False
        user = create_mock_user("1")

        self.setup_mocks(mock_get_all, mock_execute_json_query, user)

        self.assertEqual(
            pid_data_api.get_data_by_pid("mock_pid", self.mock_request),
            self.mock_query_result,
        )

    @patch.object(main_data_api, "execute_json_query")
    @patch.object(pid_path_api, "get_all")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_anonymous_user_not_public_cannot_access(
        self,
        mock_settings,
        mock_workspace_api,  # noqa, pylint: disable=unused-argument
        mock_get_all,
        mock_execute_json_query,
    ):
        """test_anonymous_user_not_public_cannot_access"""
        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = False
        user = create_mock_user("1", is_anonymous=True)

        self.setup_mocks(mock_get_all, mock_execute_json_query, user)

        with self.assertRaises(AccessControlError):
            pid_data_api.get_data_by_pid("mock_pid", self.mock_request)

    @patch.object(main_data_api, "execute_json_query")
    @patch.object(pid_path_api, "get_all")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_anonymous_user_and_public_can_access(
        self,
        mock_settings,
        mock_workspace_api,
        mock_get_all,
        mock_execute_json_query,
    ):
        """test_anonymous_user_and_public_can_access"""
        mock_public_workspace = MagicMock()
        self.mock_query_result.workspace = mock_public_workspace
        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = True
        mock_workspace_api.get_all_workspaces_with_read_access_by_user.return_value = [
            mock_public_workspace
        ]
        user = create_mock_user("1", is_anonymous=True)

        self.setup_mocks(mock_get_all, mock_execute_json_query, user)

        self.assertEqual(
            pid_data_api.get_data_by_pid("mock_pid", self.mock_request),
            self.mock_query_result,
        )


class TestGetPidForData(TestCase):
    """ACL tests for `get_pid_for_data` function."""

    def setUp(self) -> None:
        """setUp"""
        self.mock_request = MagicMock()
        self.mock_data = MagicMock()

        self.mock_data_pid = "mock_data_pid"

    def setup_mocks(
        self,
        mock_data,
        mock_get_by_id,
        mock_get_value_from_dot_notation,
        user,
        owner=None,
    ) -> None:
        """setup_mocks"""
        self.mock_request.user = user
        self.mock_data.user_id = owner.id if owner else user.id

        mock_data.get_by_id.return_value = self.mock_data
        mock_get_by_id.return_value = self.mock_data

        mock_get_value_from_dot_notation.return_value = self.mock_data_pid

    @patch.object(pid_data_api, "is_valid_pid_value")
    @patch.object(pid_data_api, "is_dot_notation_in_dictionary")
    @patch.object(pid_data_api, "get_value_from_dot_notation")
    @patch.object(pid_path_api, "get_by_template")
    @patch.object(main_data_api, "get_by_id")
    @patch.object(pid_data_acl, "Data")
    def test_superuser_can_access(
        self,
        mock_data,
        mock_get_by_id,
        mock_get_by_template,  # noqa, pylint: disable=unused-argument
        mock_get_value_from_dot_notation,
        mock_is_dot_notation_in_dictionary,  # noqa, pylint: disable=unused-argument
        mock_is_valid_pid_value,  # noqa, pylint: disable=unused-argument
    ):
        """test_superuser_can_access"""
        user = create_mock_user("1", is_superuser=True)

        self.setup_mocks(
            mock_data,
            mock_get_by_id,
            mock_get_value_from_dot_notation,
            user,
        )

        self.assertEqual(
            pid_data_api.get_pid_for_data("mock_data_id", self.mock_request),
            self.mock_data_pid,
        )

    @patch.object(pid_data_api, "is_valid_pid_value")
    @patch.object(pid_data_api, "is_dot_notation_in_dictionary")
    @patch.object(pid_data_api, "get_value_from_dot_notation")
    @patch.object(pid_path_api, "get_by_template")
    @patch.object(main_data_api, "get_by_id")
    @patch.object(pid_data_acl, "Data")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_registered_user_not_owner_cannot_access_private(
        self,
        mock_settings,
        mock_workspace_api,  # noqa, pylint: disable=unused-argument
        mock_data,
        mock_get_by_id,
        mock_get_by_template,  # noqa, pylint: disable=unused-argument
        mock_get_value_from_dot_notation,
        mock_is_dot_notation_in_dictionary,  # noqa, pylint: disable=unused-argument
        mock_is_valid_pid_value,  # noqa, pylint: disable=unused-argument
    ):
        """test_registered_user_not_owner_cannot_access_private"""
        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = False

        owner = create_mock_user("1")
        user = create_mock_user("2")

        self.setup_mocks(
            mock_data,
            mock_get_by_id,
            mock_get_value_from_dot_notation,
            user,
            owner,
        )

        with self.assertRaises(AccessControlError):
            pid_data_api.get_pid_for_data("mock_data_id", self.mock_request)

    @patch.object(pid_data_api, "is_valid_pid_value")
    @patch.object(pid_data_api, "is_dot_notation_in_dictionary")
    @patch.object(pid_data_api, "get_value_from_dot_notation")
    @patch.object(pid_path_api, "get_by_template")
    @patch.object(main_data_api, "get_by_id")
    @patch.object(pid_data_acl, "Data")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_registered_user_not_owner_can_access_public(
        self,
        mock_settings,
        mock_workspace_api,
        mock_data,
        mock_get_by_id,
        mock_get_by_template,  # noqa, pylint: disable=unused-argument
        mock_get_value_from_dot_notation,
        mock_is_dot_notation_in_dictionary,  # noqa, pylint: disable=unused-argument
        mock_is_valid_pid_value,  # noqa, pylint: disable=unused-argument
    ):
        """test_registered_user_not_owner_can_access_public"""
        mock_public_workspace = MagicMock()
        self.mock_data.workspace = mock_public_workspace
        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = True
        mock_workspace_api.get_all_workspaces_with_read_access_by_user.return_value = [
            mock_public_workspace
        ]
        user = create_mock_user("1")
        owner = create_mock_user("2")

        self.setup_mocks(
            mock_data,
            mock_get_by_id,
            mock_get_value_from_dot_notation,
            user,
            owner,
        )

        self.assertEqual(
            pid_data_api.get_pid_for_data("mock_data_id", self.mock_request),
            self.mock_data_pid,
        )

    @patch.object(pid_data_api, "is_valid_pid_value")
    @patch.object(pid_data_api, "is_dot_notation_in_dictionary")
    @patch.object(pid_data_api, "get_value_from_dot_notation")
    @patch.object(pid_path_api, "get_by_template")
    @patch.object(main_data_api, "get_by_id")
    @patch.object(pid_data_acl, "Data")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_registered_user_and_owner_can_access_private(
        self,
        mock_settings,
        mock_workspace_api,  # noqa, pylint: disable=unused-argument
        mock_data,
        mock_get_by_id,
        mock_get_by_template,  # noqa, pylint: disable=unused-argument
        mock_get_value_from_dot_notation,
        mock_is_dot_notation_in_dictionary,  # noqa, pylint: disable=unused-argument
        mock_is_valid_pid_value,  # noqa, pylint: disable=unused-argument
    ):
        """test_registered_user_and_owner_can_access_private"""
        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = False
        user = create_mock_user("1")

        self.setup_mocks(
            mock_data,
            mock_get_by_id,
            mock_get_value_from_dot_notation,
            user,
        )

        self.assertEqual(
            pid_data_api.get_pid_for_data("mock_data_id", self.mock_request),
            self.mock_data_pid,
        )

    @patch.object(pid_data_api, "is_valid_pid_value")
    @patch.object(pid_data_api, "is_dot_notation_in_dictionary")
    @patch.object(pid_data_api, "get_value_from_dot_notation")
    @patch.object(pid_path_api, "get_by_template")
    @patch.object(main_data_api, "get_by_id")
    @patch.object(pid_data_acl, "Data")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_anonymous_user_not_public_cannot_access(
        self,
        mock_settings,
        mock_workspace_api,
        mock_data,
        mock_get_by_id,
        mock_get_by_template,  # noqa, pylint: disable=unused-argument
        mock_get_value_from_dot_notation,
        mock_is_dot_notation_in_dictionary,  # noqa, pylint: disable=unused-argument
        mock_is_valid_pid_value,  # noqa, pylint: disable=unused-argument
    ):
        """test_anonymous_user_not_public_cannot_access"""
        mock_public_workspace = MagicMock()
        self.mock_data.workspace = mock_public_workspace
        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = False
        mock_workspace_api.get_all_workspaces_with_read_access_by_user.return_value = [
            mock_public_workspace
        ]
        user = create_mock_user("1", is_anonymous=True)

        self.setup_mocks(
            mock_data,
            mock_get_by_id,
            mock_get_value_from_dot_notation,
            user,
        )

        with self.assertRaises(AccessControlError):
            pid_data_api.get_pid_for_data("mock_data_id", self.mock_request)

    @patch.object(pid_data_api, "is_valid_pid_value")
    @patch.object(pid_data_api, "is_dot_notation_in_dictionary")
    @patch.object(pid_data_api, "get_value_from_dot_notation")
    @patch.object(pid_path_api, "get_by_template")
    @patch.object(main_data_api, "get_by_id")
    @patch.object(pid_data_acl, "Data")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_anonymous_user_and_public_can_access(
        self,
        mock_settings,
        mock_workspace_api,
        mock_data,
        mock_get_by_id,
        mock_get_by_template,  # noqa, pylint: disable=unused-argument
        mock_get_value_from_dot_notation,
        mock_is_dot_notation_in_dictionary,  # noqa, pylint: disable=unused-argument
        mock_is_valid_pid_value,  # noqa, pylint: disable=unused-argument
    ):
        """test_anonymous_user_and_public_can_access"""
        mock_public_workspace = MagicMock()
        self.mock_data.workspace = mock_public_workspace
        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = True
        mock_workspace_api.get_all_workspaces_with_read_access_by_user.return_value = [
            mock_public_workspace
        ]
        user = create_mock_user("1", is_anonymous=True)

        self.setup_mocks(
            mock_data,
            mock_get_by_id,
            mock_get_value_from_dot_notation,
            user,
        )

        self.assertEqual(
            pid_data_api.get_pid_for_data("mock_data_id", self.mock_request),
            self.mock_data_pid,
        )



""" Initial migration to instantiate necessary models.

Generated by Django 3.2 on 2021-10-18
"""
import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    """Migration class."""

    initial = True

    dependencies = [
        ("core_main_app", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="LocalId",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "record_name",
                    models.CharField(
                        max_length=255,
                        unique=True,
                        validators=[
                            django.core.validators.RegexValidator(
                                code="invalid_title",
                                message="Title must not be empty or only whitespaces",
                                regex=".*\\S.*",
                            )
                        ],
                    ),
                ),
                (
                    "record_object_class",
                    models.CharField(blank=True, max_length=255),
                ),
                (
                    "record_object_id",
                    models.CharField(blank=True, max_length=255),
                ),
            ],
        ),
        migrations.CreateModel(
            name="PidSettings",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("auto_set_pid", models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name="PidXpath",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("xpath", models.CharField(max_length=255)),
                (
                    "template",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="core_main_app.template",
                    ),
                ),
            ],
        ),
    ]



""" Unit tests for core_linked_records_app.components.data.watch
"""
import json
from unittest import TestCase
from unittest.mock import patch, MagicMock

from rest_framework import status

from core_linked_records_app.components.data import watch as data_watch
from core_linked_records_app.components.pid_settings.models import PidSettings
from core_linked_records_app.system.data import api as data_system_api
from core_linked_records_app.system.pid_path import (
    api as pid_path_system_api,
)
from core_linked_records_app.utils import data as data_utils
from core_linked_records_app.utils import exceptions
from core_linked_records_app.utils import providers as providers_utils
from tests import mocks


class TestRegisterPidForDataId(TestCase):
    """Unit tests for `_register_pid_for_data_id` function."""

    def setUp(self) -> None:
        """setUp"""
        self.mock_resolver_match = MagicMock()
        self.mock_resolver_match.view_name = (
            "core_linked_records_provider_record"
        )

        self.mock_provider = MagicMock()
        self.mock_provider.provider_lookup_url = "http://cdcs.handle.net"
        self.mock_provider.local_url = "http://cdcs.example.com/pid/rest"

        self.mock_prefix = "12.34567"
        self.mock_record = "mock_record"

        self.kwargs = {
            "provider_name": "mock_provider",
            "pid_value": f"{self.mock_provider.provider_lookup_url}/{self.mock_prefix}/"
            f"{self.mock_record}",
            "data_id": 42,
        }

    @patch.object(data_watch, "ProviderManager")
    def test_provider_manager_is_instantiated(self, mock_provider_manager):
        """test_provider_manager_is_instantiated"""
        with self.assertRaises(exceptions.PidResolverError):
            data_watch._register_pid_for_data_id(**self.kwargs)

        mock_provider_manager.assert_called()

    @patch.object(data_watch, "ProviderManager")
    def test_default_provider_is_retrieved(self, mock_provider_manager):
        """test_default_provider_is_retrieved"""
        with self.assertRaises(exceptions.PidResolverError):
            data_watch._register_pid_for_data_id(**self.kwargs)

        mock_provider_manager().get.assert_called_with(
            self.kwargs["provider_name"]
        )

    @patch.object(data_watch, "logger")
    @patch.object(data_watch, "resolve")
    @patch.object(data_watch, "ProviderManager")
    def test_resolver_error_is_logged_and_raises_core_error(
        self, mock_provider_manager, mock_resolve, mock_logger
    ):
        """test_resolver_error_is_logged_and_raises_core_error"""
        mock_provider_manager().get.return_value = self.mock_provider
        mock_resolve.side_effect = Exception("mock_resolve_exception")

        with self.assertRaises(exceptions.PidResolverError):
            data_watch._register_pid_for_data_id(**self.kwargs)

        mock_logger.error.assert_called()

    @patch.object(data_watch, "logger")
    @patch.object(data_watch, "resolve")
    @patch.object(data_watch, "ProviderManager")
    def test_resolver_not_matching_correct_view_is_logged_and_raises_core_error(
        self, mock_provider_manager, mock_resolve, mock_logger
    ):
        """test_resolver_not_matching_correct_view_is_logged_and_raises_core_error"""
        mock_provider_manager().get.return_value = self.mock_provider
        mock_resolve.return_value = self.mock_resolver_match
        self.mock_resolver_match.view_name = "mock_incorrect_view_name"

        with self.assertRaises(exceptions.PidResolverError):
            data_watch._register_pid_for_data_id(**self.kwargs)

        mock_logger.error.assert_called()

    @patch.object(data_watch, "logger")
    @patch.object(data_watch, "split_prefix_from_record")
    @patch.object(data_watch, "resolve")
    @patch.object(data_watch, "ProviderManager")
    def test_split_prefix_from_record_error_is_logged_and_raises_core_error(
        self,
        mock_provider_manager,
        mock_resolve,
        mock_split_prefix_from_record,
        mock_logger,
    ):
        """test_split_prefix_from_record_error_is_logged_and_raises_core_error"""
        mock_provider_manager().get.return_value = self.mock_provider
        mock_resolve.return_value = self.mock_resolver_match
        mock_split_prefix_from_record.side_effect = Exception(
            "mock_split_prefix_from_record_exception"
        )

        with self.assertRaises(exceptions.PidResolverError):
            data_watch._register_pid_for_data_id(**self.kwargs)

        mock_logger.error.assert_called()

    @patch.object(data_watch, "data_system_api")
    @patch.object(data_watch, "split_prefix_from_record")
    @patch.object(data_watch, "resolve")
    @patch.object(data_watch, "ProviderManager")
    def test_provider_create_is_called(
        self,
        mock_provider_manager,
        mock_resolve,
        mock_split_prefix_from_record,
        mock_data_system_api,  # noqa, pylint: disable=unused-argument
    ):
        """test_provider_create_is_called"""
        mock_provider_manager().get.return_value = self.mock_provider
        mock_resolve.return_value = self.mock_resolver_match
        mock_split_prefix_from_record.return_value = (
            self.mock_prefix,
            self.mock_record,
        )

        with self.assertRaises(exceptions.PidCreateError):
            data_watch._register_pid_for_data_id(**self.kwargs)

        self.mock_provider.create.assert_called_with(
            self.mock_prefix, self.mock_record
        )

    @patch.object(data_watch, "logger")
    @patch.object(data_watch, "data_system_api")
    @patch.object(data_watch, "split_prefix_from_record")
    @patch.object(data_watch, "resolve")
    @patch.object(data_watch, "ProviderManager")
    def test_provider_create_unparsable_response_is_logged_and_raises_core_error(
        self,
        mock_provider_manager,
        mock_resolve,
        mock_split_prefix_from_record,
        mock_data_system_api,  # noqa, pylint: disable=unused-argument
        mock_logger,
    ):
        """test_provider_create_unparsable_response_is_logged_and_raises_core_error"""
        mock_provider_manager().get.return_value = self.mock_provider
        mock_resolve.return_value = self.mock_resolver_match
        mock_split_prefix_from_record.return_value = (
            self.mock_prefix,
            self.mock_record,
        )

        mock_provider_create_response = MagicMock()
        mock_provider_create_response.status_code = (
            status.HTTP_500_INTERNAL_SERVER_ERROR
        )
        mock_provider_create_response.content = None
        self.mock_provider.create.return_value = mock_provider_create_response

        with self.assertRaises(exceptions.PidCreateError):
            data_watch._register_pid_for_data_id(**self.kwargs)

        mock_logger.error.assert_called()

    @patch.object(data_watch, "logger")
    @patch.object(data_watch, "data_system_api")
    @patch.object(data_watch, "split_prefix_from_record")
    @patch.object(data_watch, "resolve")
    @patch.object(data_watch, "ProviderManager")
    def test_provider_create_parsable_response_is_logged_and_raises_core_error(
        self,
        mock_provider_manager,
        mock_resolve,
        mock_split_prefix_from_record,
        mock_data_system_api,  # noqa, pylint: disable=unused-argument
        mock_logger,
    ):
        """test_provider_create_parsable_response_is_logged_and_raises_core_error"""
        mock_provider_manager().get.return_value = self.mock_provider
        mock_resolve.return_value = self.mock_resolver_match
        mock_split_prefix_from_record.return_value = (
            self.mock_prefix,
            self.mock_record,
        )

        mock_provider_create_response = MagicMock()
        mock_provider_create_response.status_code = (
            status.HTTP_500_INTERNAL_SERVER_ERROR
        )
        mock_provider_create_response.content = json.dumps(
            {"message": "mock_error"}
        )
        self.mock_provider.create.return_value = mock_provider_create_response

        with self.assertRaises(exceptions.PidCreateError):
            data_watch._register_pid_for_data_id(**self.kwargs)

        mock_logger.error.assert_called()

    @patch.object(data_watch, "data_system_api")
    @patch.object(data_watch, "split_prefix_from_record")
    @patch.object(data_watch, "resolve")
    @patch.object(data_watch, "ProviderManager")
    def test_succesful_execution_returns_record_url(
        self,
        mock_provider_manager,
        mock_resolve,
        mock_split_prefix_from_record,
        mock_data_system_api,  # noqa, pylint: disable=unused-argument
    ):
        """test_succesful_execution_returns_record_url"""
        expected_result = "mock_record_url"

        mock_provider_manager().get.return_value = self.mock_provider
        mock_resolve.return_value = self.mock_resolver_match
        mock_split_prefix_from_record.return_value = (
            self.mock_prefix,
            self.mock_record,
        )

        mock_provider_create_response = MagicMock()
        mock_provider_create_response.status_code = status.HTTP_201_CREATED
        mock_provider_create_response.content = json.dumps(
            {"url": expected_result}
        )
        self.mock_provider.create.return_value = mock_provider_create_response

        result = data_watch._register_pid_for_data_id(**self.kwargs)

        self.assertEqual(result, expected_result)


class TestSetDataPid(TestCase):
    """Unit tests for `_set_data_pid` function."""

    def setUp(self):
        """setUp"""
        self.mock_kwargs = {"instance": mocks.MockData()}

    @patch.object(data_watch, "transaction")
    def test_transaction_on_commit_is_called(self, mock_transaction):
        """test_transaction_on_commit_is_called"""
        data_watch.set_data_pid(None, self.mock_kwargs["instance"])
        mock_transaction.on_commit.assert_called()

    @patch.object(PidSettings, "get")
    def test_pid_settings_get_failure_raises_pid_create_error(
        self, mock_pid_settings_get
    ):
        """test_pid_settings_get_failure_raises_core_error"""

        mock_pid_settings_get.side_effect = Exception(
            "mock_pid_settings_get_exception"
        )

        with self.assertRaises(exceptions.PidCreateError):
            data_watch._set_data_pid(**self.mock_kwargs)

    @patch.object(pid_path_system_api, "get_pid_path_by_template")
    @patch.object(PidSettings, "get")
    def test_get_pid_path_by_template_failure_raises_pid_create_error(
        self, mock_pid_settings_get, mock_get_pid_path_by_template
    ):
        """test_get_pid_path_by_template_failure_raises_pid_create_error"""

        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_path_by_template.side_effect = Exception(
            "mock_get_pid_path_by_template_exception"
        )

        with self.assertRaises(exceptions.PidCreateError):
            data_watch._set_data_pid(**self.mock_kwargs)

    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(pid_path_system_api, "get_pid_path_by_template")
    @patch.object(PidSettings, "get")
    def test_get_pid_value_for_data_failure_returns_none(
        self,
        mock_pid_settings_get,
        mock_get_pid_path_by_template,
        mock_get_pid_value_for_data,
    ):
        """test_get_pid_value_for_data_failure_returns_none"""

        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_path_by_template.return_value = mocks.MockPidPath()
        mock_get_pid_value_for_data.side_effect = Exception(
            "mock_get_pid_value_for_data_exception"
        )

        self.assertIsNone(data_watch._set_data_pid(**self.mock_kwargs))

    @patch.object(data_system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(pid_path_system_api, "get_pid_path_by_template")
    @patch.object(PidSettings, "get")
    def test_delete_pid_for_data_raises_pid_create_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_path_by_template,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
    ):
        """test_delete_pid_for_data_raises_pid_create_error"""

        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_path_by_template.return_value = mocks.MockPidPath()
        mock_get_pid_value_for_data.return_value = "mock_pid_value"
        mock_delete_pid_for_data.side_effect = Exception(
            "mock_delete_pid_for_data_exception"
        )

        with self.assertRaises(exceptions.PidCreateError):
            data_watch._set_data_pid(**self.mock_kwargs)

    @patch.object(providers_utils, "retrieve_provider_name")
    @patch.object(data_system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(pid_path_system_api, "get_pid_path_by_template")
    @patch.object(PidSettings, "get")
    def test_retrieve_provider_name_failure_raise_pid_create_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_path_by_template,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
        mock_retrieve_provider_name,
    ):
        """test_retrieve_provider_name_failure_raise_pid_create_error"""

        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_path_by_template.return_value = mocks.MockPidPath()
        mock_get_pid_value_for_data.return_value = "mock_pid_value"
        mock_delete_pid_for_data.return_value = None
        mock_retrieve_provider_name.side_effect = Exception(
            "mock_retrieve_provider_name_exception"
        )

        with self.assertRaises(exceptions.PidCreateError):
            data_watch._set_data_pid(**self.mock_kwargs)

    @patch.object(providers_utils.ProviderManager, "get")
    @patch.object(providers_utils, "retrieve_provider_name")
    @patch.object(data_system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(pid_path_system_api, "get_pid_path_by_template")
    @patch.object(PidSettings, "get")
    def test_provider_manager_get_failure_raise_pid_create_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_path_by_template,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
        mock_retrieve_provider_name,
        mock_provider_manager_get,
    ):
        """test_provider_manager_get_failure_raise_pid_create_error"""

        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_path_by_template.return_value = mocks.MockPidPath()
        mock_get_pid_value_for_data.return_value = None
        mock_delete_pid_for_data.return_value = None
        mock_retrieve_provider_name.return_value = "mock_provider_name"
        mock_provider_manager_get.side_effect = Exception(
            "mock_provider_manager_get_exception"
        )

        with self.assertRaises(exceptions.PidCreateError):
            data_watch._set_data_pid(**self.mock_kwargs)

    @patch.object(data_system_api, "is_pid_defined")
    @patch.object(providers_utils.ProviderManager, "get")
    @patch.object(providers_utils, "retrieve_provider_name")
    @patch.object(data_system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(pid_path_system_api, "get_pid_path_by_template")
    @patch.object(PidSettings, "get")
    def test_is_pid_defined_failure_raise_pid_create_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_path_by_template,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
        mock_retrieve_provider_name,
        mock_provider_manager_get,
        mock_is_pid_defined,
    ):
        """test_is_pid_defined_failure_raise_pid_create_error"""

        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_path_by_template.return_value = mocks.MockPidPath()
        mock_get_pid_value_for_data.return_value = None
        mock_delete_pid_for_data.return_value = None
        mock_retrieve_provider_name.return_value = "mock_provider_name"
        mock_provider_manager_get.return_value = mocks.MockProviderManager()
        mock_is_pid_defined.side_effect = Exception(
            "mock_provider_manager_get_exception"
        )

        with self.assertRaises(exceptions.PidCreateError):
            data_watch._set_data_pid(**self.mock_kwargs)

    @patch.object(data_system_api, "is_pid_defined_for_data")
    @patch.object(data_system_api, "is_pid_defined")
    @patch.object(providers_utils.ProviderManager, "get")
    @patch.object(providers_utils, "retrieve_provider_name")
    @patch.object(data_system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(pid_path_system_api, "get_pid_path_by_template")
    @patch.object(PidSettings, "get")
    def test_is_pid_defined_for_data_failure_raise_pid_create_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_path_by_template,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
        mock_retrieve_provider_name,
        mock_provider_manager_get,
        mock_is_pid_defined,
        mock_is_pid_defined_for_data,
    ):
        """test_is_pid_defined_for_data_failure_raise_pid_create_error"""

        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_path_by_template.return_value = mocks.MockPidPath()
        mock_get_pid_value_for_data.return_value = None
        mock_delete_pid_for_data.return_value = None
        mock_retrieve_provider_name.return_value = "mock_provider_name"
        mock_provider_manager_get.return_value = mocks.MockProviderManager()
        mock_is_pid_defined.return_value = True
        mock_is_pid_defined_for_data.side_effect = Exception(
            "mock_is_pid_defined_for_data_exception"
        )

        with self.assertRaises(exceptions.PidCreateError):
            data_watch._set_data_pid(**self.mock_kwargs)

    @patch.object(data_system_api, "is_pid_defined_for_data")
    @patch.object(data_system_api, "is_pid_defined")
    @patch.object(providers_utils.ProviderManager, "get")
    @patch.object(providers_utils, "retrieve_provider_name")
    @patch.object(data_system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(pid_path_system_api, "get_pid_path_by_template")
    @patch.object(PidSettings, "get")
    def test_pid_already_defined_raise_pid_create_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_path_by_template,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
        mock_retrieve_provider_name,
        mock_provider_manager_get,
        mock_is_pid_defined,
        mock_is_pid_defined_for_data,
    ):
        """test_pid_already_defined_raise_pid_create_error"""

        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_path_by_template.return_value = mocks.MockPidPath()
        mock_get_pid_value_for_data.return_value = None
        mock_delete_pid_for_data.return_value = None
        mock_retrieve_provider_name.return_value = "mock_provider_name"
        mock_provider_manager_get.return_value = mocks.MockProviderManager()
        mock_is_pid_defined.return_value = True
        mock_is_pid_defined_for_data.return_value = False

        with self.assertRaises(exceptions.PidCreateError):
            data_watch._set_data_pid(**self.mock_kwargs)

    @patch.object(data_watch, "_register_pid_for_data_id")
    @patch.object(data_system_api, "is_pid_defined_for_data")
    @patch.object(data_system_api, "is_pid_defined")
    @patch.object(providers_utils.ProviderManager, "get")
    @patch.object(providers_utils, "retrieve_provider_name")
    @patch.object(data_system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(pid_path_system_api, "get_pid_path_by_template")
    @patch.object(PidSettings, "get")
    def test_register_pid_for_data_id_failure_raise_pid_create_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_path_by_template,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
        mock_retrieve_provider_name,
        mock_provider_manager_get,
        mock_is_pid_defined,
        mock_is_pid_defined_for_data,
        mock_register_pid_for_data_id,
    ):
        """test_register_pid_for_data_id_failure_raise_pid_create_error"""

        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_path_by_template.return_value = mocks.MockPidPath()
        mock_get_pid_value_for_data.return_value = None
        mock_delete_pid_for_data.return_value = None
        mock_retrieve_provider_name.return_value = "mock_provider_name"
        mock_provider_manager_get.return_value = mocks.MockProviderManager()
        mock_is_pid_defined.return_value = True
        mock_is_pid_defined_for_data.return_value = True
        mock_register_pid_for_data_id.side_effect = Exception(
            "mock_register_pid_for_data_id_exception"
        )

        with self.assertRaises(exceptions.PidCreateError):
            data_watch._set_data_pid(**self.mock_kwargs)

    @patch.object(data_utils, "set_pid_value_for_data")
    @patch.object(data_watch, "_register_pid_for_data_id")
    @patch.object(data_system_api, "is_pid_defined_for_data")
    @patch.object(data_system_api, "is_pid_defined")
    @patch.object(providers_utils.ProviderManager, "get")
    @patch.object(providers_utils, "retrieve_provider_name")
    @patch.object(data_system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(pid_path_system_api, "get_pid_path_by_template")
    @patch.object(PidSettings, "get")
    def test_set_pid_value_for_data_failure_raise_pid_create_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_path_by_template,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
        mock_retrieve_provider_name,
        mock_provider_manager_get,
        mock_is_pid_defined,
        mock_is_pid_defined_for_data,
        mock_register_pid_for_data_id,
        mock_set_pid_value_for_data,
    ):
        """test_set_pid_value_for_data_failure_raise_pid_create_error"""

        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_path_by_template.return_value = mocks.MockPidPath()
        mock_get_pid_value_for_data.return_value = None
        mock_delete_pid_for_data.return_value = None
        mock_retrieve_provider_name.return_value = "mock_provider_name"
        mock_provider_manager_get.return_value = mocks.MockProviderManager()
        mock_is_pid_defined.return_value = True
        mock_is_pid_defined_for_data.return_value = True
        mock_register_pid_for_data_id.return_value = "mock_pid_value"
        mock_set_pid_value_for_data.side_effect = Exception(
            "mock_set_pid_value_for_data_exception"
        )

        with self.assertRaises(exceptions.PidCreateError):
            data_watch._set_data_pid(**self.mock_kwargs)

    @patch.object(data_utils, "set_pid_value_for_data")
    @patch.object(data_watch, "_register_pid_for_data_id")
    @patch.object(data_system_api, "is_pid_defined_for_data")
    @patch.object(data_system_api, "is_pid_defined")
    @patch.object(providers_utils.ProviderManager, "get")
    @patch.object(providers_utils, "retrieve_provider_name")
    @patch.object(data_system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(pid_path_system_api, "get_pid_path_by_template")
    @patch.object(PidSettings, "get")
    def test_default_execution_returns_none(
        self,
        mock_pid_settings_get,
        mock_get_pid_path_by_template,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
        mock_retrieve_provider_name,
        mock_provider_manager_get,
        mock_is_pid_defined,
        mock_is_pid_defined_for_data,
        mock_register_pid_for_data_id,
        mock_set_pid_value_for_data,
    ):
        """test_default_execution_returns_none"""

        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_path_by_template.return_value = mocks.MockPidPath()
        mock_get_pid_value_for_data.return_value = None
        mock_delete_pid_for_data.return_value = None
        mock_retrieve_provider_name.return_value = "mock_provider_name"
        mock_provider_manager_get.return_value = mocks.MockProviderManager()
        mock_is_pid_defined.return_value = True
        mock_is_pid_defined_for_data.return_value = True
        mock_register_pid_for_data_id.return_value = "mock_pid_value"
        mock_set_pid_value_for_data.return_value = None

        self.assertIsNone(data_watch._set_data_pid(**self.mock_kwargs))


class TestDeleteDataPid(TestCase):
    """Unit tests for `detele_data_pid` function."""

    def setUp(self) -> None:
        """setUp"""
        self.mock_data = mocks.MockDocument()

    @patch.object(data_system_api, "delete_pid_for_data")
    def test_delete_pid_for_data_is_called(self, mock_delete_pid_for_data):
        """test_delete_pid_for_data_is_called"""
        data_watch.delete_data_pid(None, self.mock_data)

        mock_delete_pid_for_data.assert_called_with(self.mock_data)

    @patch.object(data_watch, "logger")
    @patch.object(data_system_api, "delete_pid_for_data")
    def test_delete_pid_for_data_error_raise_warning(
        self, mock_delete_pid_for_data, mock_logger
    ):
        """test_delete_pid_for_data_error_raise_warning"""
        mock_delete_pid_for_data.side_effect = Exception(
            "mock_delete_pid_for_data_exception"
        )
        data_watch.delete_data_pid(None, self.mock_data)

        mock_logger.warning.assert_called()



""" Unit tests for `core_linked_records_app.utils.xml`.
"""
from unittest import TestCase
from unittest.mock import patch, MagicMock, call
from core_linked_records_app.utils import xml as linked_records_xml_utils
from xml_utils.commons import exceptions as xml_utils_exceptions


class TestGetValueAtXPath(TestCase):
    """Unit tests for `get_value_at_xpath` function."""

    def setUp(self):
        """setUp"""
        self.mock_kwargs = {
            "xml_tree": MagicMock(),
            "xpath": MagicMock(),
            "namespaces": MagicMock(),
        }

    def test_xml_tree_xpath_called(self):
        """test_xml_tree_xpath_called"""
        with self.assertRaises(xml_utils_exceptions.XPathError):
            linked_records_xml_utils.get_value_at_xpath(**self.mock_kwargs)

        self.mock_kwargs["xml_tree"].xpath.assert_called_with(
            self.mock_kwargs["xpath"],
            namespaces=self.mock_kwargs["namespaces"],
        )

    def test_more_than_one_xpath_found_raises_xpath_error(self):
        """test_more_than_one_xpath_found_raises_xpath_error"""
        self.mock_kwargs["xml_tree"].xpath.return_value = [MagicMock()] * 2

        with self.assertRaises(xml_utils_exceptions.XPathError):
            linked_records_xml_utils.get_value_at_xpath(**self.mock_kwargs)

    def test_no_xpath_found_raises_xpath_error(self):
        """test_no_xpath_found_raises_xpath_error"""
        self.mock_kwargs["xml_tree"].xpath.return_value = []

        with self.assertRaises(xml_utils_exceptions.XPathError):
            linked_records_xml_utils.get_value_at_xpath(**self.mock_kwargs)

    def test_retrieve_xpath_element_text(self):
        """test_retrieve_xpath_element_text"""
        mock_xpath_element = MagicMock()
        self.mock_kwargs["xml_tree"].xpath.return_value = [mock_xpath_element]

        self.assertEqual(
            linked_records_xml_utils.get_value_at_xpath(**self.mock_kwargs),
            str(mock_xpath_element.text),
        )

    def test_retrieve_xpath_element_if_text_raises_attribute_error(self):
        """test_retrieve_xpath_element_if_text_raises_attribute_error"""
        mock_xpath_element = "mock_xpath"
        self.mock_kwargs["xml_tree"].xpath.return_value = [mock_xpath_element]

        self.assertEqual(
            linked_records_xml_utils.get_value_at_xpath(**self.mock_kwargs),
            str(mock_xpath_element),
        )

    def test_retrieve_xpath_element_none_returns_none(self):
        """test_retrieve_xpath_element_none_returns_none"""
        mock_xpath_element = None
        self.mock_kwargs["xml_tree"].xpath.return_value = [mock_xpath_element]

        self.assertIsNone(
            linked_records_xml_utils.get_value_at_xpath(**self.mock_kwargs)
        )


class TestCanCreateValueAtXpath(TestCase):
    """Unit tests for `can_create_value_at_xpath` function."""

    def setUp(self):
        """setUp"""
        self.mock_kwargs = {
            "xml_string": MagicMock(),
            "xsd_string": MagicMock(),
            "xpath": MagicMock(),
            "value": MagicMock(),
        }

    @patch.object(linked_records_xml_utils, "validate_xml_data")
    @patch.object(linked_records_xml_utils, "set_value_at_xpath")
    @patch.object(linked_records_xml_utils, "create_tree_from_xpath")
    @patch.object(linked_records_xml_utils, "XSDTree")
    @patch.object(
        linked_records_xml_utils, "get_target_namespace_for_xsd_string"
    )
    def test_get_target_namespace_for_xsd_string_called(
        self,
        mock_get_target_namespace_for_xsd_string,
        mock_xsd_tree,  # noqa, pylint: disable=unused-argument
        mock_create_tree_from_xpath,  # noqa, pylint: disable=unused-argument
        mock_set_value_at_xpath,  # noqa, pylint: disable=unused-argument
        mock_validate_xml_data,  # noqa, pylint: disable=unused-argument
    ):
        """test_get_target_namespace_for_xsd_string_called"""
        linked_records_xml_utils.can_create_value_at_xpath(**self.mock_kwargs)

        mock_get_target_namespace_for_xsd_string.assert_called_with(
            self.mock_kwargs["xsd_string"]
        )

    @patch.object(linked_records_xml_utils, "validate_xml_data")
    @patch.object(linked_records_xml_utils, "set_value_at_xpath")
    @patch.object(linked_records_xml_utils, "create_tree_from_xpath")
    @patch.object(linked_records_xml_utils, "XSDTree")
    @patch.object(
        linked_records_xml_utils, "get_target_namespace_for_xsd_string"
    )
    def test_xsd_tree_build_tree_called(
        self,
        mock_get_target_namespace_for_xsd_string,  # noqa, pylint: disable=unused-argument
        mock_xsd_tree,
        mock_create_tree_from_xpath,  # noqa, pylint: disable=unused-argument
        mock_set_value_at_xpath,  # noqa, pylint: disable=unused-argument
        mock_validate_xml_data,  # noqa, pylint: disable=unused-argument
    ):
        """test_xsd_tree_build_tree_called"""
        linked_records_xml_utils.can_create_value_at_xpath(**self.mock_kwargs)

        mock_xsd_tree.build_tree.assert_has_calls(
            [
                call(self.mock_kwargs["xml_string"]),
                call(self.mock_kwargs["xsd_string"]),
            ]
        )

    @patch.object(linked_records_xml_utils, "validate_xml_data")
    @patch.object(linked_records_xml_utils, "set_value_at_xpath")
    @patch.object(linked_records_xml_utils, "create_tree_from_xpath")
    @patch.object(linked_records_xml_utils, "XSDTree")
    @patch.object(
        linked_records_xml_utils, "get_target_namespace_for_xsd_string"
    )
    def test_create_tree_from_xpath_called(
        self,
        mock_get_target_namespace_for_xsd_string,
        mock_xsd_tree,
        mock_create_tree_from_xpath,
        mock_set_value_at_xpath,  # noqa, pylint: disable=unused-argument
        mock_validate_xml_data,  # noqa, pylint: disable=unused-argument
    ):
        """test_create_tree_from_xpath_called"""
        mock_target_namespace = MagicMock()
        mock_xml_tree = MagicMock()

        mock_get_target_namespace_for_xsd_string.return_value = (
            mock_target_namespace
        )
        mock_xsd_tree.build_tree.return_value = mock_xml_tree

        linked_records_xml_utils.can_create_value_at_xpath(**self.mock_kwargs)

        mock_create_tree_from_xpath.assert_called_with(
            self.mock_kwargs["xpath"],
            mock_xml_tree,
            mock_target_namespace,
        )

    @patch.object(linked_records_xml_utils, "validate_xml_data")
    @patch.object(linked_records_xml_utils, "set_value_at_xpath")
    @patch.object(linked_records_xml_utils, "create_tree_from_xpath")
    @patch.object(linked_records_xml_utils, "XSDTree")
    @patch.object(
        linked_records_xml_utils, "get_target_namespace_for_xsd_string"
    )
    def test_set_value_at_xpath_called(
        self,
        mock_get_target_namespace_for_xsd_string,
        mock_xsd_tree,  # noqa, pylint: disable=unused-argument
        mock_create_tree_from_xpath,
        mock_set_value_at_xpath,
        mock_validate_xml_data,  # noqa, pylint: disable=unused-argument
    ):
        """test_set_value_at_xpath_called"""
        mock_target_namespace = MagicMock()
        mock_modified_xml_tree = MagicMock()

        mock_get_target_namespace_for_xsd_string.return_value = (
            mock_target_namespace
        )
        mock_create_tree_from_xpath.return_value = mock_modified_xml_tree

        linked_records_xml_utils.can_create_value_at_xpath(**self.mock_kwargs)

        mock_set_value_at_xpath.assert_called_with(
            mock_modified_xml_tree,
            self.mock_kwargs["xpath"],
            self.mock_kwargs["value"],
            mock_target_namespace,
        )

    @patch.object(linked_records_xml_utils, "validate_xml_data")
    @patch.object(linked_records_xml_utils, "set_value_at_xpath")
    @patch.object(linked_records_xml_utils, "create_tree_from_xpath")
    @patch.object(linked_records_xml_utils, "XSDTree")
    @patch.object(
        linked_records_xml_utils, "get_target_namespace_for_xsd_string"
    )
    def test_validate_xml_data_called(
        self,
        mock_get_target_namespace_for_xsd_string,  # noqa, pylint: disable=unused-argument
        mock_xsd_tree,
        mock_create_tree_from_xpath,
        mock_set_value_at_xpath,  # noqa, pylint: disable=unused-argument
        mock_validate_xml_data,
    ):
        """test_validate_xml_data_called"""
        mock_xsd_built_tree = MagicMock()
        mock_modified_xml_tree = MagicMock()

        mock_xsd_tree.build_tree.return_value = mock_xsd_built_tree
        mock_create_tree_from_xpath.return_value = mock_modified_xml_tree

        linked_records_xml_utils.can_create_value_at_xpath(**self.mock_kwargs)

        mock_validate_xml_data.assert_called_with(
            mock_xsd_built_tree,
            mock_modified_xml_tree,
        )

    @patch.object(linked_records_xml_utils, "validate_xml_data")
    @patch.object(linked_records_xml_utils, "set_value_at_xpath")
    @patch.object(linked_records_xml_utils, "create_tree_from_xpath")
    @patch.object(linked_records_xml_utils, "XSDTree")
    @patch.object(
        linked_records_xml_utils, "get_target_namespace_for_xsd_string"
    )
    def test_validate_xml_data_not_none_returns_false(
        self,
        mock_get_target_namespace_for_xsd_string,  # noqa, pylint: disable=unused-argument
        mock_xsd_tree,  # noqa, pylint: disable=unused-argument
        mock_create_tree_from_xpath,  # noqa, pylint: disable=unused-argument
        mock_set_value_at_xpath,  # noqa, pylint: disable=unused-argument
        mock_validate_xml_data,
    ):
        """test_validate_xml_data_not_none_returns_false"""
        mock_validate_xml_data.return_value = "mock_validate_xml_error"

        self.assertFalse(
            linked_records_xml_utils.can_create_value_at_xpath(
                **self.mock_kwargs
            )
        )

    @patch.object(linked_records_xml_utils, "validate_xml_data")
    @patch.object(linked_records_xml_utils, "set_value_at_xpath")
    @patch.object(linked_records_xml_utils, "create_tree_from_xpath")
    @patch.object(linked_records_xml_utils, "XSDTree")
    @patch.object(
        linked_records_xml_utils, "get_target_namespace_for_xsd_string"
    )
    def test_successful_execution_returns_true(
        self,
        mock_get_target_namespace_for_xsd_string,  # noqa, pylint: disable=unused-argument
        mock_xsd_tree,  # noqa, pylint: disable=unused-argument
        mock_create_tree_from_xpath,  # noqa, pylint: disable=unused-argument
        mock_set_value_at_xpath,  # noqa, pylint: disable=unused-argument
        mock_validate_xml_data,
    ):
        """test_successful_execution_returns_true"""
        mock_validate_xml_data.return_value = None

        self.assertTrue(
            linked_records_xml_utils.can_create_value_at_xpath(
                **self.mock_kwargs
            )
        )



""" REST views for the query API
"""
from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from core_linked_records_app.utils.query import execute_local_pid_query


class RetrieveQueryPidListView(APIView):
    """Retrieve list of PIDs given a query"""

    def post(self, request):
        try:
            return Response(
                execute_local_pid_query(request.data, request),
                status=status.HTTP_200_OK,
            )
        except Exception as exc:
            return Response(
                f"An unexpected error occured while retrieving PID list: {str(exc)}",
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )



""" Unit tests for admin views
"""
from unittest import TestCase
from unittest.mock import patch

from core_linked_records_app.views.admin.views import PidSettingsView
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from tests.mocks import MockRequest, MockTemplate, MockPidPath


class TestPidSettingsViewGet(TestCase):
    """Tests PidSettingsView get method"""

    def setUp(self) -> None:
        self.view = PidSettingsView()
        mock_user = create_mock_user(1, is_superuser=True)
        self.request = MockRequest()
        self.request.user = mock_user
        self.settings_json = {
            "path": "mock_path",
            "system_name": "mock_system_name",
            "prefixes": ["mock_prefix"],
        }

    @patch("core_linked_records_app.views.admin.views.admin_render")
    @patch("core_linked_records_app.views.admin.views.pid_path_api")
    def test_pid_path_api_get_all_called(
        self,
        mock_get_pid_path_api,
        mock_admin_render,
    ):
        """test_pid_path_api_get_all_called"""
        mock_admin_render.return_value = None

        self.view.get(self.request)
        mock_get_pid_path_api.get_all.assert_called_with(self.request)

    @patch("core_linked_records_app.views.admin.views.admin_render")
    @patch("core_linked_records_app.views.admin.views.pid_path_api")
    def test_pid_path_get_all_failure_returns_error_page(
        self,
        mock_get_pid_path_api,
        mock_admin_render,
    ):
        """test_pid_path_get_all_failure_returns_error_page"""
        mock_get_pid_path_api.get_all.side_effect = Exception(
            "mock_get_pid_path_api_get_all_exception"
        )

        self.view.get(self.request)
        mock_admin_render.assert_called_with(
            self.request,
            "core_linked_records_app/admin/pid_settings_error.html",
            context={
                "error": "An error occured while retrieving the PID settings. "
                "Please contact an administrator for more information. Exception: "
                "mock_get_pid_path_api_get_all_exception."
            },
        )

    @patch("core_linked_records_app.views.admin.views.admin_render")
    @patch("core_linked_records_app.views.admin.views.get_pid_settings_dict")
    @patch("core_linked_records_app.views.admin.views.pid_path_api")
    def test_get_pid_settings_dict_called(
        self,
        mock_get_pid_path_api,
        mock_get_pid_settings_dict,
        mock_admin_render,
    ):
        """test_get_pid_settings_dictcalled"""
        mock_get_pid_path_api.get_all.return_value = []
        mock_admin_render.return_value = None

        self.view.get(self.request)
        mock_get_pid_settings_dict.assert_called()

    @patch("core_linked_records_app.views.admin.views.admin_render")
    @patch("core_linked_records_app.views.admin.views.get_pid_settings_dict")
    @patch("core_linked_records_app.views.admin.views.pid_path_api")
    def test_get_pid_settings_dict_failure_returns_error_page(
        self,
        mock_get_pid_path_api,
        mock_get_pid_settings_dict,
        mock_admin_render,
    ):
        """test_get_pid_settings_dict_failure_returns_error_page"""
        mock_get_pid_path_api.get_all.return_value = []
        mock_get_pid_settings_dict.side_effect = Exception(
            "mock_get_pid_settings_dict_exception"
        )
        mock_admin_render.return_value = None

        self.view.get(self.request)
        mock_admin_render.assert_called_with(
            self.request,
            "core_linked_records_app/admin/pid_settings_error.html",
            context={
                "error": "An error occured while retrieving the PID settings. "
                "Please contact an administrator for more information. Exception: "
                "mock_get_pid_settings_dict_exception."
            },
        )

    @patch("core_linked_records_app.views.admin.views.admin_render")
    @patch("core_linked_records_app.views.admin.views.get_pid_settings_dict")
    @patch("core_linked_records_app.views.admin.views.pid_path_api")
    def test_return_settings_page(
        self,
        mock_get_pid_path_api,
        mock_get_pid_settings_dict,
        mock_admin_render,
    ):
        """test_return_settings_page"""
        mock_get_pid_path_api.get_all.return_value = []
        mock_get_pid_settings_dict.return_value = self.settings_json
        mock_admin_render.return_value = None

        self.view.get(self.request)
        mock_admin_render.assert_called_with(
            self.request,
            "core_linked_records_app/admin/pid_settings.html",
            assets={
                "js": [
                    {
                        "path": "core_linked_records_app/admin/js/pid_settings/auto_set_pid.js",
                        "is_raw": False,
                    }
                ],
                "css": [
                    "core_linked_records_app/admin/css/pid_settings.css",
                ],
            },
            context={"pid_settings": self.settings_json},
        )

    @patch("core_linked_records_app.views.admin.views.admin_render")
    @patch("core_linked_records_app.views.admin.views.get_pid_settings_dict")
    @patch("core_linked_records_app.views.admin.views.pid_path_api")
    def test_return_settings_page_with_pid_path(
        self,
        mock_get_pid_path_api,
        mock_get_pid_settings_dict,
        mock_admin_render,
    ):
        """test_return_settings_page_with_pid_path"""
        mock_pid_path1 = MockPidPath()
        mock_pid_path1.id = 1
        mock_pid_path1.path = "root.elem1"
        mock_pid_path1.template = MockTemplate(display_name="Template01")

        mock_pid_path2 = MockPidPath()
        mock_pid_path2.id = 2
        mock_pid_path2.path = "root.elem2"
        mock_pid_path2.template = MockTemplate(display_name="Template02")

        mock_get_pid_path_api.get_all.return_value = [
            mock_pid_path1,
            mock_pid_path2,
        ]
        mock_get_pid_settings_dict.return_value = self.settings_json
        mock_admin_render.return_value = None

        self.view.get(self.request)
        mock_admin_render.assert_called_with(
            self.request,
            "core_linked_records_app/admin/pid_settings.html",
            assets={
                "js": [
                    {
                        "path": "core_linked_records_app/admin/js/pid_settings/auto_set_pid.js",
                        "is_raw": False,
                    }
                ],
                "css": [
                    "core_linked_records_app/admin/css/pid_settings.css",
                ],
            },
            context={"pid_settings": self.settings_json},
        )



""" Unit tests for PidPath admin site
"""
from unittest import TestCase
from unittest.mock import Mock

from core_linked_records_app.components.pid_path.admin_site import (
    CustomPidPathAdmin,
)


class TestCustomPidPathAdminHasAddPermission(TestCase):
    """Test has_add_permission in CustomPidPathAdmin view."""

    def test_can_add_item(self):
        """test_can_add_item"""
        custom_pid_path_model_admin = Mock(spec=CustomPidPathAdmin)
        self.assertTrue(
            custom_pid_path_model_admin.has_add_permission(None, None)
        )



""" Data Xml renderer for django REST API
"""
from django.http import HttpResponse
from rest_framework import renderers
from rest_framework.status import (
    HTTP_404_NOT_FOUND,
    HTTP_500_INTERNAL_SERVER_ERROR,
)
from core_linked_records_app.settings import (
    BACKWARD_COMPATIBILITY_DATA_XML_CONTENT,
)


class DataXmlRenderer(renderers.BaseRenderer):
    """Data Xml Renderer"""

    media_type = "application/xml"
    format = "xml"
    charset = "utf-8"

    def render(self, data, media_type=None, renderer_context=None):
        """Render the data object by just returning the xml_content field

        Args:
            data:
            media_type:
            renderer_context:

        Returns: xml string
        """

        # check errors
        if "status" in data and data["status"] == "error":
            if data["code"] == HTTP_404_NOT_FOUND:
                return HttpResponse(status=HTTP_404_NOT_FOUND)

            return HttpResponse(status=HTTP_500_INTERNAL_SERVER_ERROR)

        return data[
            "xml_content"
            if BACKWARD_COMPATIBILITY_DATA_XML_CONTENT
            else "content"
        ]



""" Migration to create necessary permissions for API functions.

Generated by Django 4.2.3 on 2023-08-17
"""

from django.db import migrations, models


class Migration(migrations.Migration):
    """Migration class."""

    dependencies = [
        ("core_linked_records_app", "0002_alter_pidxpath_xpath"),
    ]

    operations = [
        migrations.CreateModel(
            name="LinkedRecords",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "verbose_name": "core_linked_records_app",
                "permissions": (
                    ("read_pid_settings", "Can read pid settings"),
                ),
                "default_permissions": (),
            },
        ),
    ]



""" Linked Records Settings
"""
from django.core.exceptions import ObjectDoesNotExist
from django.db import models

from core_main_app.commons import exceptions


class PidSettings(models.Model):
    """Pid Settings"""

    auto_set_pid = models.BooleanField(default=False)

    @staticmethod
    def get():
        """Retrieve the PidSettings.

        Returns:
             PidSettings - first PidSettings object
        """
        try:
            return PidSettings.objects.first()  # pylint: disable=no-member
        except ObjectDoesNotExist:
            return None
        except Exception as exc:
            raise exceptions.ModelError(str(exc))

    def __str__(self):
        """LocalId object as string.

        Returns:
            str - String representation of PidSettings object.
        """
        return (
            f"PidSettings {{ id:{self.pk}; auto_set_pid:{self.auto_set_pid} }}"
        )



""" Unit tests for `core_linked_records_app.access_control.discover`.
"""
from unittest import TestCase
from unittest.mock import patch, MagicMock, call

from core_linked_records_app.access_control import discover
from core_linked_records_app.access_control import rights
from core_main_app.permissions import rights as main_rights


class TestInitPermissions(TestCase):
    """Unit tests for `init_permission` function."""

    def setUp(self) -> None:
        """setUp"""
        self.mock_default_group = MagicMock()
        self.mock_anonymous_group = MagicMock()

    @patch.object(discover, "logger")
    @patch.object(discover, "permissions_api")
    @patch.object(discover, "group_api")
    def test_groups_retrieval_error_is_logged(
        self,
        mock_group_api,
        mock_permissions_api,  # noqa, pylint: disable=unused-argument
        mock_logger,
    ):
        """test_groups_retrieval_error_is_logged"""
        mock_group_api.get_or_create.side_effect = Exception(
            "mock_group_exception"
        )
        discover.init_permissions()
        mock_logger.error.assert_called()

    @patch.object(discover, "permissions_api")
    @patch.object(discover, "group_api")
    def test_groups_are_retrieved(
        self,
        mock_group_api,
        mock_permissions_api,  # noqa, pylint: disable=unused-argument
    ):
        """test_groups_are_retrieved"""
        mock_group_api.get_or_create.return_value = (MagicMock(), False)
        discover.init_permissions()
        mock_group_api.get_or_create.assert_has_calls(
            [
                call(name=main_rights.DEFAULT_GROUP),
            ]
        )

    @patch.object(discover, "logger")
    @patch.object(discover, "permissions_api")
    @patch.object(discover, "group_api")
    def test_permissions_retrieval_error_is_logged(
        self,
        mock_group_api,  # noqa, pylint: disable=unused-argument
        mock_permissions_api,
        mock_logger,
    ):
        """test_permissions_retrieval_error_is_logged"""
        mock_permissions_api.get_by_codename.side_effect = Exception(
            "mock_get_by_codename_exception"
        )
        discover.init_permissions()
        mock_logger.error.assert_called()

    @patch.object(discover, "permissions_api")
    @patch.object(discover, "group_api")
    def test_permissions_are_retrieved(
        self,
        mock_group_api,
        mock_permissions_api,
    ):
        """test_permissions_are_retrieved"""
        mock_group_api.get_or_create.return_value = (MagicMock(), False)
        discover.init_permissions()
        mock_permissions_api.get_by_codename.assert_has_calls(
            [
                call(codename=rights.CAN_READ_PID_SETTINGS),
            ]
        )

    @patch.object(discover, "logger")
    @patch.object(discover, "permissions_api")
    @patch.object(discover, "group_api")
    def test_default_group_add_permissions_error_is_logged(
        self,
        mock_group_api,
        mock_permissions_api,  # noqa, pylint: disable=unused-argument
        mock_logger,
    ):
        """test_default_group_add_permissions_error_is_logged"""
        self.mock_default_group.permissions.add.side_effect = Exception(
            "mock_default_group_permission_add_exception"
        )
        mock_group_api.get_or_create.side_effect = [
            (self.mock_default_group, False),
        ]

        discover.init_permissions()
        mock_logger.error.assert_called()

    @patch.object(discover, "permissions_api")
    @patch.object(discover, "group_api")
    def test_default_group_has_all_permissions(
        self,
        mock_group_api,
        mock_permissions_api,
    ):
        """test_default_group_has_all_permissions"""
        mock_permission_object_list = ["can_read_pid_settings_perm"]

        mock_group_api.get_or_create.side_effect = [
            (self.mock_default_group, False),
            (self.mock_anonymous_group, False),
        ]
        mock_permissions_api.get_by_codename.side_effect = (
            mock_permission_object_list
        )

        discover.init_permissions()
        self.mock_default_group.permissions.add.assert_called_with(
            *mock_permission_object_list
        )



""" Handle.net implementation class
"""
import json
import logging
from base64 import b64encode

from core_linked_records_app import settings
from core_linked_records_app.utils.providers import AbstractIdProvider
from core_main_app.utils.requests_utils.requests_utils import (
    send_put_request,
    send_delete_request,
    send_get_request,
)

logger = logging.getLogger(__name__)


class HandleNetSystem(AbstractIdProvider):
    """Provider implemented using Handle.Net."""

    registration_api = "api/handles"
    response_code_messages = {
        1: "Successful operation",
        2: "Unexpected handle server error",
        100: "Handle not found",
        101: "Handle already exists",
        102: "Invalid handle",
        200: "Values not found",
        201: "Value already exists",
        202: "Invalid value",
        301: "Server not responsible for handle",
        402: "Authentication needed",
    }

    def __init__(
        self,
        provider_name,
        provider_lookup_url,
        provider_registration_url,
        username,
        password,
    ):
        self.provider_registration_url = provider_registration_url
        self.auth_token = b64encode(
            f"{username}:{password}".encode("utf-8")
        ).decode("utf-8")
        super().__init__(provider_name, provider_lookup_url)

    def _get_message_for_response_code(self, return_code):
        if return_code in self.response_code_messages.keys():
            return self.response_code_messages[return_code]

        return "Handle code not recognized"

    def _update_response_content(self, response):
        json_response_content = json.loads(response.content)

        json_response_content["message"] = self._get_message_for_response_code(
            json_response_content["responseCode"]
        )

        json_response_content[
            "url"
        ] = f'{self.provider_lookup_url}/{ json_response_content["handle"]}'

        return json.dumps(json_response_content)

    def _generate_record_data(self, record, include_handle=False):
        record_data = {
            "values": [
                {
                    "index": settings.HANDLE_NET_RECORD_INDEX,
                    "type": "URL",
                    "data": {
                        "format": "string",
                        "value": f"{self.local_url}/{record}",
                    },
                },
                settings.HANDLE_NET_ADMIN_DATA,
            ]
        }

        if include_handle:
            record_data["handle"] = record

        return json.dumps(record_data)

    def get(self, record):
        """Retrieve an existring handle.net handle.

        Args:
            record:

        Returns:
        """
        response = send_get_request(
            f"{self.provider_lookup_url}/{self.registration_api}/{record}",
            headers={
                "Content-Type": "application/json",
            },
        )

        response._content = self._update_response_content(response)
        return response

    def create(self, prefix, record=None):
        """Create a new handle for a handle.net system.

        Args:
            prefix:
            record:

        Returns:
        """
        # Create request url depending on handle value
        if record is not None:
            request_url = f"{self.provider_registration_url}/{self.registration_api}/{prefix}/{record}?overwrite=false"

        else:
            request_url = f"{self.provider_registration_url}/{self.registration_api}/{prefix}/?overwrite=false&mintNewSuffix=true"

        response = send_put_request(
            request_url,
            self._generate_record_data(f"{prefix}/{record}"),
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Basic {str(self.auth_token)}",
            },
        )

        if record is None:
            return self.update(response.json()["handle"])

        response._content = self._update_response_content(response)
        return response

    def update(self, record):
        """Update a handle for a handle.net system.

        Args:
            record:

        Returns:
        """
        response = send_put_request(
            f"{self.provider_registration_url}/{self.registration_api}/"
            f"{record}?overwrite=true",
            self._generate_record_data(record, include_handle=True),
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Basic {str(self.auth_token)}",
            },
        )

        response._content = self._update_response_content(response)
        return response

    def delete(self, record):
        response = send_delete_request(
            f"{self.provider_registration_url}/{self.registration_api}/{record}",
            headers={"Authorization": f"Basic {str(self.auth_token)}"},
        )

        response._content = self._update_response_content(response)
        return response



""" Protected API to manage PidSettings object.
"""
import logging

from core_linked_records_app.components.pid_settings.access_control import (
    can_get_pid_settings,
    can_upsert_pid_settings,
)
from core_linked_records_app.system.pid_settings import (
    api as pid_settings_system_api,
)
from core_main_app.access_control.decorators import access_control

logger = logging.getLogger(__name__)


@access_control(can_upsert_pid_settings)
def upsert(pid_settings_object, user):  # noqa, pylint: disable=unused-argument
    """Insert or update the PidSettings provided as parameter.

    Args:
        pid_settings_object:

    Returns:
        Saved PidSettings object
    """
    return pid_settings_system_api.upsert(pid_settings_object)


@access_control(can_get_pid_settings)
def get(user):  # noqa, pylint: disable=unused-argument
    """Retrieve the PidSettings object from DB.

    Returns:
        PidSettings object
    """
    return pid_settings_system_api.get()



""" Unit tests for core_linked_records_app.system.pid_settings.api
"""

from unittest import TestCase
from unittest.mock import patch

from core_linked_records_app.components.pid_settings import (
    access_control as pid_settings_acl,
)
from core_linked_records_app.components.pid_settings.models import PidSettings
from core_linked_records_app.system.pid_settings import (
    api as pid_settings_system_api,
)
from core_main_app.commons.exceptions import ApiError


class TestUpsert(TestCase):
    """Unit tests for `upsert` function."""

    def setUp(self) -> None:
        """setUp"""
        self.mock_pid_settings = PidSettings()

    @patch.object(PidSettings, "save")
    def test_save_failure_raises_api_error(self, mock_save):
        """test_save_failure_raises_api_error"""

        mock_save.side_effect = Exception("mock_save_exception")

        with self.assertRaises(ApiError):
            pid_settings_system_api.upsert(self.mock_pid_settings)

    @patch.object(PidSettings, "save")
    def test_returns_save_method_output(self, mock_save):
        """test_returns_save_method_output"""

        mock_save.return_value = None

        self.assertEqual(
            pid_settings_system_api.upsert(self.mock_pid_settings),
            self.mock_pid_settings,
        )


class TestGet(TestCase):
    """Unit tests for `get` function."""

    @patch.object(PidSettings, "get")
    @patch.object(pid_settings_acl, "check_has_perm")
    def test_save_failure_raises_api_error(
        self, mock_check_has_perm, mock_get
    ):
        """test_save_failure_raises_api_error"""
        mock_check_has_perm.return_value = True
        mock_get.side_effect = Exception("mock_get_exception")

        with self.assertRaises(ApiError):
            pid_settings_system_api.get()

    @patch.object(PidSettings, "get")
    @patch.object(pid_settings_acl, "check_has_perm")
    def test_returns_save_method_output(self, mock_check_has_perm, mock_get):
        """test_returns_save_method_output"""
        mock_check_has_perm.return_value = True
        expected_result = "mock_get"
        mock_get.return_value = expected_result

        self.assertEqual(pid_settings_system_api.get(), expected_result)



""" Unit tests for `core_linked_records_app.components.blob.api`.
"""
from unittest import TestCase
from unittest.mock import patch, MagicMock

from core_linked_records_app.components.blob import access_control as blob_acl
from core_linked_records_app.components.blob import api as blob_api
from core_linked_records_app.components.local_id.models import LocalId
from core_linked_records_app.system.local_id import api as local_id_system_api
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.commons import exceptions
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from tests import mocks


class TestGetBlobByPid(TestCase):
    """Unit tests for `get_blob_by_pid` function."""

    def setUp(self):
        """setUp"""
        self.user = create_mock_user("1")

    @patch.object(local_id_system_api, "get_by_name")
    def test_get_by_name_error_raises_api_error(self, mock_get_by_name):
        """test_get_by_name_error_raises_api_error"""

        mock_get_by_name.side_effect = Exception("mock_get_by_name_exception")
        mock_valid_pid = "https://websi.te/provider/record"

        with self.assertRaises(exceptions.ApiError):
            blob_api.get_blob_by_pid(mock_valid_pid, self.user)

    @patch.object(local_id_system_api, "get_by_name")
    def test_undefined_classpath_raises_does_not_exist(self, mock_get_by_name):
        """test_undefined_classpath_raises_does_not_exist"""

        mock_valid_pid = "https://websi.te/provider/record"
        mock_get_by_name.return_value = LocalId(record_name=mock_valid_pid)

        with self.assertRaises(exceptions.DoesNotExist):
            blob_api.get_blob_by_pid(mock_valid_pid, self.user)

    @patch.object(blob_api, "import_module")
    @patch.object(local_id_system_api, "get_by_name")
    def test_failed_import_raises_api_error(
        self,
        mock_get_by_name,
        mock_import_module,
    ):
        """test_failed_import_raises_api_error"""
        mock_valid_pid = "https://websi.te/provider/record"
        mock_get_by_name.return_value = LocalId(
            record_name=mock_valid_pid,
            record_object_class="mock_record_object_class",
            record_object_id="mock_record_object_id",
        )
        mock_import_module.side_effect = Exception(
            "mock_import_module_exception"
        )

        with self.assertRaises(exceptions.ApiError):
            blob_api.get_blob_by_pid(mock_valid_pid, self.user)

    @patch.object(blob_api, "getattr")
    @patch.object(blob_api, "import_module")
    @patch.object(local_id_system_api, "get_by_name")
    def test_get_by_id_exception_raises_api_error(
        self,
        mock_get_by_name,
        mock_import_module,
        mock_getattr,
    ):
        """test_get_by_id_exception_raises_api_error"""
        mock_valid_pid = "https://websi.te/provider/record"
        mock_get_by_name.return_value = LocalId(
            record_name=mock_valid_pid,
            record_object_class="mock_record_object_class",
            record_object_id="mock_record_object_id",
        )
        mock_import_module.return_value = None
        mock_getattr.side_effect = Exception("mock_getattr_exception")

        with self.assertRaises(exceptions.ApiError):
            blob_api.get_blob_by_pid(mock_valid_pid, self.user)

    @patch.object(blob_api, "getattr")
    @patch.object(blob_api, "import_module")
    @patch.object(local_id_system_api, "get_by_name")
    def test_get_by_id_acl_error_raises_acl_error(
        self,
        mock_get_by_name,
        mock_import_module,
        mock_getattr,
    ):
        """test_get_by_id_acl_error_raises_acl_error"""
        mock_valid_pid = "https://websi.te/provider/record"
        mock_get_by_name.return_value = LocalId(
            record_name=mock_valid_pid,
            record_object_class="mock_record_object_class",
            record_object_id="mock_record_object_id",
        )
        mock_import_module.return_value = None
        mock_getattr.side_effect = AccessControlError(
            "mock_getattr_access_control_error"
        )

        with self.assertRaises(AccessControlError):
            blob_api.get_blob_by_pid(mock_valid_pid, self.user)

    @patch.object(blob_api, "getattr")
    @patch.object(blob_api, "import_module")
    @patch.object(local_id_system_api, "get_by_name")
    def test_returns_expected_get_by_id_output(
        self,
        mock_get_by_name,
        mock_import_module,
        mock_getattr,
    ):
        """test_returns_expected_get_by_id_output"""
        mock_valid_pid = "https://websi.te/provider/record"
        mock_get_by_name.return_value = LocalId(
            record_name=mock_valid_pid,
            record_object_class="mock_record_object_class",
            record_object_id="mock_record_object_id",
        )
        mock_import_module.return_value = None

        mock_blob = mocks.MockDocument()
        mock_blob.user_id = self.user.id
        mock_blob_module = mocks.MockModule()
        mock_blob_module.get_by_id_return_value = mock_blob
        mock_getattr.return_value = mock_blob_module

        result = blob_api.get_blob_by_pid(mock_valid_pid, self.user)
        self.assertEqual(result, mock_blob)


class TestGetPidForBlob(TestCase):
    """Unit tests for `get_pid_for_blob` function."""

    @patch.object(blob_api, "blob_system_api")
    @patch.object(blob_acl, "check_can_read_document")
    @patch.object(blob_acl, "Blob")
    def test_get_pid_for_blob_called(
        self,
        mock_blob,
        mock_check_can_read_document,
        mock_blob_system_api,
    ):
        """test_get_pid_for_blob_called"""
        mock_blob_id = "mock_blob_id"
        mock_blob.get_by_id.return_value = MagicMock()
        mock_check_can_read_document.return_value = True

        blob_api.get_pid_for_blob(mock_blob_id, create_mock_user("1"))
        mock_blob_system_api.get_pid_for_blob.assert_called_with(mock_blob_id)

    @patch.object(blob_api, "blob_system_api")
    @patch.object(blob_acl, "check_can_read_document")
    @patch.object(blob_acl, "Blob")
    def test_success_returns_system_get_pid_for_blob_output(
        self,
        mock_blob,
        mock_check_can_read_document,
        mock_blob_system_api,
    ):
        """test_success_returns_system_get_pid_for_blob_output"""
        mock_blob_id = "mock_blob_id"
        mock_blob.get_by_id.return_value = MagicMock()
        mock_check_can_read_document.return_value = True

        expected_result = "mock_blob_pid"
        mock_blob_system_api.get_pid_for_blob.return_value = expected_result

        self.assertEqual(
            blob_api.get_pid_for_blob(mock_blob_id, create_mock_user("1")),
            expected_result,
        )


class TestSetPidForBlob(TestCase):
    """Unit tests for `set_pid_for_blob` function."""

    @patch.object(blob_api, "blob_system_api")
    @patch.object(blob_acl, "check_can_read_document")
    @patch.object(blob_acl, "Blob")
    def test_set_pid_for_blob_called(
        self,
        mock_blob,
        mock_check_can_read_document,
        mock_blob_system_api,
    ):
        """test_set_pid_for_blob_called"""
        mock_blob_id = "mock_blob_id"
        mock_blob_pid = "mock_blob_pid"
        mock_blob.get_by_id.return_value = MagicMock()
        mock_check_can_read_document.return_value = True

        blob_api.set_pid_for_blob(
            mock_blob_id, mock_blob_pid, create_mock_user("1")
        )
        mock_blob_system_api.set_pid_for_blob.assert_called_with(
            mock_blob_id, mock_blob_pid
        )

    @patch.object(blob_api, "blob_system_api")
    @patch.object(blob_acl, "check_can_read_document")
    @patch.object(blob_acl, "Blob")
    def test_success_returns_system_set_pid_for_blob_output(
        self,
        mock_blob,
        mock_check_can_read_document,
        mock_blob_system_api,
    ):
        """test_success_returns_system_set_pid_for_blob_output"""
        mock_blob_id = "mock_blob_id"
        mock_blob_pid = "mock_blob_pid"
        mock_blob.get_by_id.return_value = MagicMock()
        mock_check_can_read_document.return_value = True

        expected_results = "mock_blob_pid"
        mock_blob_system_api.set_pid_for_blob.return_value = expected_results

        self.assertEqual(
            blob_api.set_pid_for_blob(
                mock_blob_id, mock_blob_pid, create_mock_user("1")
            ),
            expected_results,
        )



""" Ajax views accessible by users.
"""
import json
from urllib.parse import urljoin

from django.urls import reverse
from django.utils import timezone
from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from core_explore_common_app.commons.exceptions import ExploreRequestError
from core_explore_common_app.components.query import api as query_api
from core_explore_common_app.utils.oaipmh import oaipmh as oaipmh_utils
from core_explore_common_app.utils.protocols.oauth2 import (
    send_post_request as oauth2_post_request,
    send_get_request as oauth2_get_request,
)
from core_explore_common_app.utils.query import query as query_utils
from core_linked_records_app import settings
from core_linked_records_app.components.blob import api as blob_api
from core_linked_records_app.components.data import api as data_api
from core_linked_records_app.utils.query import execute_local_pid_query

if (
    "core_oaipmh_harvester_app" in settings.INSTALLED_APPS
    and "core_explore_oaipmh_app" in settings.INSTALLED_APPS
):  # Import OAI-PMH pid views if packages are present.
    from core_linked_records_app.utils.query import (
        execute_oaipmh_pid_query,
    )


class RetrieveDataPIDView(APIView):
    """Retrieve PIDs for a given data IDs."""

    def get(self, request):
        """get

        Args:
            request:

        Returns:
        """
        try:
            if "data_id" in request.GET:  # Local data
                return Response(
                    {
                        "pid": data_api.get_pid_for_data(
                            request.GET["data_id"], request
                        )
                    }
                )
            if (
                "core_oaipmh_harvester_app" in settings.INSTALLED_APPS
                and "core_explore_oaipmh_app" in settings.INSTALLED_APPS
                and "oai_data_id" in request.GET
            ):  # OAI-PMH data
                from core_linked_records_app.components.oai_record import (
                    api as oai_record_api,
                )

                return Response(
                    {
                        "pid": oai_record_api.get_pid_for_data(
                            request.GET["oai_data_id"], request
                        )
                    }
                )
            if (
                "core_federated_search_app" in settings.INSTALLED_APPS
                and "fede_data_id" in request.GET
                and "fede_origin" in request.GET
            ):  # Federated data
                from core_federated_search_app.components.instance import (
                    api as instance_api,
                )

                fede_origin_keys = request.GET["fede_origin"].split("&")
                instance_name = fede_origin_keys[1].split("=")[1]
                instance = instance_api.get_by_name(instance_name)

                reverse_url = reverse("core_linked_records_retrieve_data_pid")
                url_get_data = (
                    f'{reverse_url}?data_id={request.GET["fede_data_id"]}'
                )

                data_response = oauth2_get_request(
                    urljoin(instance.endpoint, url_get_data),
                    instance.access_token,
                )
                return Response(json.loads(data_response.text))

            return Response(
                {
                    "message": "Impossible to retrieve PID for data with the given "
                    "parameters"
                },
                status=status.HTTP_400_BAD_REQUEST,
            )
        except Exception as exc:
            return Response(
                {
                    "message": f"An unexpected exception occurred while retrieving data "
                    f"PID: {str(exc)}"
                },
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )


class RetrieveBlobPIDView(APIView):
    """Retrieve PIDs for a given blob ID."""

    def get(self, request):
        """get PIDs
        Args:
            request:

        Returns:

        """
        if "blob_id" in request.GET:
            try:
                blob_pid = blob_api.get_pid_for_blob(
                    request.GET["blob_id"], request.user
                )
                sub_url = reverse(
                    "core_linked_records_provider_record",
                    kwargs={
                        "provider": settings.ID_PROVIDER_SYSTEM_NAME,
                        "record": "",
                    },
                )

                return Response(
                    {
                        "pid": f"{settings.SERVER_URI}{sub_url}{blob_pid.record_name}"
                    }
                )
            except Exception as exc:
                return Response(
                    {
                        "message": f"An unexpected exception occurred while retrieving "
                        f"blob PID: {str(exc)}"
                    },
                    status=status.HTTP_500_INTERNAL_SERVER_ERROR,
                )
        else:
            return Response(
                {"message": "Missing parameter 'blob_id'."},
                status=status.HTTP_400_BAD_REQUEST,
            )


class RetrieveListPIDView(APIView):
    """Retrieve PIDs for a given list of data IDs."""

    def post(self, request):
        """Retrieve PIDs
        Args:
            request:

        Returns:

        """
        try:
            # FIXME duplicated code with core_explore_common.utils.query.send
            query = query_api.get_by_id(
                request.data.get("query_id", None),
                request.user,
            )
            data_source = query.data_sources[
                int(request.data.get("data_source_index", 0))
            ]

            # Build serialized query to send to data source
            json_query = {
                "query": query.content,
                "templates": json.dumps(
                    [
                        {"id": template.id, "hash": template.hash}
                        for template in query.templates.all()
                    ]
                ),
                "options": json.dumps(data_source["query_options"]),
                "order_by_field": data_source["order_by_field"],
            }

            if (
                data_source["authentication"]["auth_type"] == "session"
            ):  # Local and OAI-PMH data sources
                if query_utils.is_local_data_source(data_source):
                    json_response = execute_local_pid_query(
                        json_query, request
                    )
                elif oaipmh_utils.is_oai_data_source(data_source):

                    json_response = execute_oaipmh_pid_query(
                        json_query, request
                    )
                else:
                    raise ExploreRequestError("Unknown data source type.")
            elif (
                data_source["authentication"]["auth_type"] == "oauth2"
            ):  # Federated data sources
                response = oauth2_post_request(
                    data_source["capabilities"]["query_pid"],
                    json_query,
                    data_source["authentication"]["params"]["access_token"],
                    session_time_zone=timezone.get_current_timezone(),
                )

                if response.status_code == status.HTTP_200_OK:
                    json_response = response.json()
                else:
                    return Response(
                        {
                            "error": f"Data source returned HTTP {response.status_code}."
                        },
                        status=status.HTTP_400_BAD_REQUEST,
                    )
            else:
                return Response(
                    {"error": "Unknown authentication type."},
                    status=status.HTTP_400_BAD_REQUEST,
                )

            return Response(
                {"pids": [pid for pid in json_response if pid is not None]},
                status=status.HTTP_200_OK,
            )
        except Exception as exception:
            return Response(
                {"error": str(exception)},
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )



""" Unit tests for core_linked_records_app.rest.providers.views
"""
import json
from unittest import TestCase
from unittest.mock import patch

from core_linked_records_app import settings
from core_linked_records_app.rest.providers import views as providers_views
from core_linked_records_app.utils.providers import ProviderManager
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.commons.exceptions import DoesNotExist
from core_main_app.rest.data.serializers import DataSerializer
from tests import mocks


class TestProviderRecordViewPost(TestCase):
    """Test Provider Record View Post"""

    def setUp(self) -> None:
        self.mock_request = mocks.MockRequest()

    def test_invalid_prefix_returns_500(self):
        """test_invalid_prefix_returns_500"""

        test_view = providers_views.ProviderRecordView()
        response = test_view.post(
            self.mock_request, "mock_provider", "mock_prefix/mock_record"
        )

        self.assertEqual(response.status_code, 500)

    def test_invalid_record_returns_500(self):
        """test_invalid_record_returns_500"""

        test_view = providers_views.ProviderRecordView()
        response = test_view.post(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/@@@@@",
        )

        self.assertEqual(response.status_code, 500)

    @patch.object(ProviderManager, "get")
    def test_provider_manager_get_fails_returns_500(
        self, mock_provider_manager_get
    ):
        """test_provider_manager_get_fails_returns_500"""

        mock_provider_manager_get.side_effect = Exception(
            "mock_provider_manager_get_exception"
        )
        test_view = providers_views.ProviderRecordView()
        response = test_view.post(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 500)

    @patch.object(ProviderManager, "get")
    def test_provider_manager_create_fails_returns_500(
        self, mock_provider_manager_get
    ):
        """test_provider_manager_create_fails_returns_500"""

        mock_provider_manager_get.return_value = mocks.MockProviderManager(
            create_exc=Exception("mock_provider_manager_create_exception")
        )
        test_view = providers_views.ProviderRecordView()
        response = test_view.post(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 500)

    @patch.object(ProviderManager, "get")
    def test_success_returns_200(self, mock_provider_manager_get):
        """test_success_returns_200"""

        mock_provider_manager_get.return_value = mocks.MockProviderManager(
            create_result=mocks.MockResponse()
        )
        test_view = providers_views.ProviderRecordView()
        response = test_view.post(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 200)


class TestProviderRecordViewPut(TestCase):
    """Test Provider Record View Put"""

    def setUp(self) -> None:
        self.mock_request = mocks.MockRequest()

    @patch.object(ProviderManager, "get")
    def test_provider_manager_get_fails_returns_500(
        self, mock_provider_manager_get
    ):
        """test_provider_manager_get_fails_returns_500"""

        mock_provider_manager_get.side_effect = Exception(
            "mock_provider_manager_get_exception"
        )
        test_view = providers_views.ProviderRecordView()
        response = test_view.put(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 500)

    @patch.object(ProviderManager, "get")
    def test_provider_manager_update_fails_returns_500(
        self, mock_provider_manager_get
    ):
        """test_provider_manager_update_fails_returns_500"""

        mock_provider_manager_get.return_value = mocks.MockProviderManager(
            update_exc=Exception("mock_provider_manager_update_exception")
        )

        test_view = providers_views.ProviderRecordView()
        response = test_view.put(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 500)

    @patch.object(ProviderManager, "get")
    def test_success_returns_200(self, mock_provider_manager_get):
        """test_success_returns_200"""

        mock_provider_manager_get.return_value = mocks.MockProviderManager(
            update_result=mocks.MockResponse()
        )

        test_view = providers_views.ProviderRecordView()
        response = test_view.put(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 200)


class TestProviderRecordViewGet(TestCase):
    """Test Provider Record View Get"""

    def setUp(self) -> None:
        self.mock_request = mocks.MockRequest()

    @patch.object(ProviderManager, "get")
    def test_provider_manager_get_fails_returns_500(
        self, mock_provider_manager_get
    ):
        """test_provider_manager_get_fails_returns_500"""

        mock_provider_manager_get.side_effect = Exception(
            "mock_provider_manager_get_exception"
        )
        test_view = providers_views.ProviderRecordView()
        response = test_view.get(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 500)

    @patch.object(ProviderManager, "get")
    def test_provider_manager_get_record_fails_returns_500(
        self, mock_provider_manager_get
    ):
        """test_provider_manager_get_record_fails_returns_500"""

        mock_provider_manager_get.return_value = mocks.MockProviderManager(
            get_exc=Exception("mock_provider_manager_get_exception")
        )
        test_view = providers_views.ProviderRecordView()
        response = test_view.get(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 500)

    @patch.object(providers_views, "get_data_by_pid")
    @patch.object(ProviderManager, "get")
    def test_get_data_by_pid_fails_returns_500(
        self, mock_provider_manager_get, mock_get_data_by_pid
    ):
        """test_get_data_by_pid_fails_returns_500"""

        mock_provider_manager_get.return_value = mocks.MockProviderManager(
            get_result=mocks.MockResponse(
                content=json.dumps({"url": "mock_url"})
            )
        )
        mock_get_data_by_pid.side_effect = Exception(
            "mock_get_data_by_pid_exception"
        )

        test_view = providers_views.ProviderRecordView()
        response = test_view.get(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 500)

    @patch.object(DataSerializer, "__new__")
    @patch.object(providers_views, "get_data_by_pid")
    @patch.object(ProviderManager, "get")
    def test_data_serializer_fails_returns_500(
        self,
        mock_provider_manager_get,
        mock_get_data_by_pid,
        mock_data_serializer,
    ):
        """test_data_serializer_fails_returns_500"""

        mock_provider_manager_get.return_value = mocks.MockProviderManager(
            get_result=mocks.MockResponse(
                content=json.dumps({"url": "mock_url"})
            )
        )
        mock_get_data_by_pid.return_value = mocks.MockData()
        mock_data_serializer.side_effect = Exception(
            "mock_data_serializer_exception"
        )

        test_view = providers_views.ProviderRecordView()
        response = test_view.get(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 500)

    @patch.object(DataSerializer, "__new__")
    @patch.object(providers_views, "get_data_by_pid")
    @patch.object(ProviderManager, "get")
    def test_get_data_success_returns_200(
        self,
        mock_provider_manager_get,
        mock_get_data_by_pid,
        mock_data_serializer,
    ):
        """test_get_data_success_returns_200"""

        mock_provider_manager_get.return_value = mocks.MockProviderManager(
            get_result=mocks.MockResponse(
                content=json.dumps({"url": "mock_url"})
            )
        )
        mock_get_data_by_pid.side_effect = mocks.MockData()
        mock_data_serializer.return_value = mocks.MockSerializer()

        test_view = providers_views.ProviderRecordView()
        response = test_view.get(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 200)

    @patch.object(providers_views, "get_blob_by_pid")
    @patch.object(providers_views, "get_data_by_pid")
    @patch.object(ProviderManager, "get")
    def test_get_blob_by_pid_fails_returns_500(
        self,
        mock_provider_manager_get,
        mock_get_data_by_pid,
        mock_get_blob_by_pid,
    ):
        """test_get_blob_by_pid_fails_returns_500"""

        mock_provider_manager_get.return_value = mocks.MockProviderManager(
            get_result=mocks.MockResponse(
                content=json.dumps({"url": "mock_url"})
            )
        )
        mock_get_data_by_pid.side_effect = DoesNotExist(
            "mock_get_data_by_pid_does_not_exist"
        )
        mock_get_blob_by_pid.side_effect = Exception(
            "mock_get_blob_by_pid_exception"
        )

        test_view = providers_views.ProviderRecordView()
        response = test_view.get(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 500)

    @patch.object(providers_views, "get_blob_by_pid")
    @patch.object(providers_views, "get_data_by_pid")
    @patch.object(ProviderManager, "get")
    def test_get_blob_by_pid_access_control_error_returns_403(
        self,
        mock_provider_manager_get,
        mock_get_data_by_pid,
        mock_get_blob_by_pid,
    ):
        """test_get_blob_by_pid_access_control_error_returns_403"""

        mock_provider_manager_get.return_value = mocks.MockProviderManager(
            get_result=mocks.MockResponse(
                content=json.dumps({"url": "mock_url"})
            )
        )
        mock_get_data_by_pid.side_effect = DoesNotExist(
            "mock_get_data_by_pid_does_not_exist"
        )
        mock_get_blob_by_pid.side_effect = AccessControlError(
            "mock_get_blob_by_pid_access_control_error"
        )

        test_view = providers_views.ProviderRecordView()
        response = test_view.get(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 403)

    @patch.object(providers_views, "get_blob_by_pid")
    @patch.object(providers_views, "get_data_by_pid")
    @patch.object(ProviderManager, "get")
    def test_get_blob_by_pid_deos_not_exist_returns_404(
        self,
        mock_provider_manager_get,
        mock_get_data_by_pid,
        mock_get_blob_by_pid,
    ):
        """test_get_blob_by_pid_deos_not_exist_returns_404"""

        mock_provider_manager_get.return_value = mocks.MockProviderManager(
            get_result=mocks.MockResponse(
                content=json.dumps({"url": "mock_url"})
            )
        )
        mock_get_data_by_pid.side_effect = DoesNotExist(
            "mock_get_data_by_pid_does_not_exist"
        )
        mock_get_blob_by_pid.side_effect = DoesNotExist(
            "mock_get_blob_by_pid_does_not_exist"
        )

        test_view = providers_views.ProviderRecordView()
        response = test_view.get(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 404)

    @patch.object(providers_views, "get_file_http_response")
    @patch.object(providers_views, "get_blob_by_pid")
    @patch.object(providers_views, "get_data_by_pid")
    @patch.object(ProviderManager, "get")
    def test_get_file_http_response_fails_returns_500(
        self,
        mock_provider_manager_get,
        mock_get_data_by_pid,
        mock_get_blob_by_pid,
        mock_get_file_http_response,
    ):
        """test_get_file_http_response_fails_returns_500"""

        mock_provider_manager_get.return_value = mocks.MockProviderManager(
            get_result=mocks.MockResponse(
                content=json.dumps({"url": "mock_url"})
            )
        )
        mock_get_data_by_pid.side_effect = DoesNotExist(
            "mock_get_data_by_pid_does_not_exist"
        )
        mock_get_blob_by_pid.return_value = mocks.MockBlob()
        mock_get_file_http_response.side_effect = Exception(
            "mock_get_file_http_response"
        )

        test_view = providers_views.ProviderRecordView()
        response = test_view.get(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 500)

    @patch.object(providers_views, "get_file_http_response")
    @patch.object(providers_views, "get_blob_by_pid")
    @patch.object(providers_views, "get_data_by_pid")
    @patch.object(ProviderManager, "get")
    def test_get_file_http_response_success_returns_200(
        self,
        mock_provider_manager_get,
        mock_get_data_by_pid,
        mock_get_blob_by_pid,
        mock_get_file_http_response,
    ):
        """test_get_file_http_response_success_returns_200"""

        mock_provider_manager_get.return_value = mocks.MockProviderManager(
            get_result=mocks.MockResponse(
                content=json.dumps({"url": "mock_url"})
            )
        )
        mock_get_data_by_pid.side_effect = DoesNotExist(
            "mock_get_data_by_pid_does_not_exist"
        )
        mock_get_blob_by_pid.return_value = mocks.MockBlob()
        mock_get_file_http_response.return_value = mocks.MockResponse()

        test_view = providers_views.ProviderRecordView()
        response = test_view.get(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 200)


class TestProviderRecordViewDelete(TestCase):
    """Test Provider Record View Delete"""

    def setUp(self) -> None:
        self.mock_request = mocks.MockRequest()

    @patch.object(ProviderManager, "get")
    def test_provider_manager_get_fails_returns_500(
        self, mock_provider_manager_get
    ):
        """test_provider_manager_get_fails_returns_500"""

        mock_provider_manager_get.side_effect = Exception(
            "mock_provider_manager_get_exception"
        )
        test_view = providers_views.ProviderRecordView()
        response = test_view.delete(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 500)

    @patch.object(ProviderManager, "get")
    def test_provider_manager_update_fails_returns_500(
        self, mock_provider_manager_get
    ):
        """test_provider_manager_update_fails_returns_500"""

        mock_provider_manager_get.return_value = mocks.MockProviderManager(
            delete_exc=Exception("mock_provider_manager_update_exception")
        )

        test_view = providers_views.ProviderRecordView()
        response = test_view.delete(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 500)

    @patch.object(ProviderManager, "get")
    def test_success_returns_200(self, mock_provider_manager_get):
        """test_success_returns_200"""

        mock_provider_manager_get.return_value = mocks.MockProviderManager(
            delete_result=mocks.MockResponse()
        )

        test_view = providers_views.ProviderRecordView()
        response = test_view.delete(
            self.mock_request,
            "mock_provider",
            f"{settings.ID_PROVIDER_PREFIXES[0]}/mock_record",
        )

        self.assertEqual(response.status_code, 200)



""" Unit tests for `core_linked_records.components.blob.access_control`.
"""
from unittest import TestCase
from unittest.mock import MagicMock, patch

from core_linked_records_app.components.blob import access_control as blob_acl
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.utils.tests_tools.MockUser import create_mock_user


class TestCanGetBlobByPid(TestCase):
    """Unit tests for `can_get_blob_by_pid` function."""

    def setUp(self) -> None:
        """setUp"""
        self.mock_func = MagicMock()
        self.mock_func_retval = "mock_blob"

        self.mock_func.return_value = self.mock_func_retval
        self.mock_kwargs = {
            "func": self.mock_func,
            "pid": "mock_pid_value",
            "user": create_mock_user("1"),
        }

    @patch.object(blob_acl, "check_can_read_document")
    def test_input_function_called(
        self,
        mock_check_can_read_document,  # noqa, pylint: disable=unused-argument
    ):
        """test_input_function_called"""
        blob_acl.can_get_blob_by_pid(**self.mock_kwargs)
        self.mock_func.assert_called_with(
            self.mock_kwargs["pid"], self.mock_kwargs["user"]
        )

    @patch.object(blob_acl, "check_can_read_document")
    def test_check_can_read_document_not_called_for_superuser(
        self,
        mock_check_can_read_document,
    ):
        """test_check_can_read_document_not_called_for_superuser"""
        self.mock_kwargs["user"] = create_mock_user("1", is_superuser=True)
        blob_acl.can_get_blob_by_pid(**self.mock_kwargs)
        mock_check_can_read_document.assert_not_called()

    @patch.object(blob_acl, "check_can_read_document")
    def test_check_can_read_document_called(
        self,
        mock_check_can_read_document,
    ):
        """test_check_can_read_document_called"""
        blob_acl.can_get_blob_by_pid(**self.mock_kwargs)
        mock_check_can_read_document.assert_called_with(
            self.mock_func_retval, self.mock_kwargs["user"]
        )

    @patch.object(blob_acl, "check_can_read_document")
    def test_check_can_read_document_exception_fails(
        self,
        mock_check_can_read_document,
    ):
        """test_check_can_read_document_exception_fails"""
        mock_check_can_read_document.side_effect = AccessControlError(
            "mock_check_can_read_document_acl_error"
        )

        with self.assertRaises(AccessControlError):
            blob_acl.can_get_blob_by_pid(**self.mock_kwargs)

    @patch.object(blob_acl, "check_can_read_document")
    def test_success_returns_function_output(
        self,
        mock_check_can_read_document,  # noqa, pylint: disable=unused-argument
    ):
        """test_success_returns_function_output"""
        self.assertEqual(
            blob_acl.can_get_blob_by_pid(**self.mock_kwargs),
            self.mock_func_retval,
        )


class TestCanGetPidForBlob(TestCase):
    """Unit tests for `can_get_pid_for_blob` function."""

    def setUp(self) -> None:
        """setUp"""
        self.mock_blob_object = "mock_blob_object"
        self.mock_func_retval = "mock_func_retval"

        mock_func = MagicMock()
        mock_func.return_value = self.mock_func_retval

        self.mock_kwargs = {
            "func": mock_func,
            "blob_id": "mock_blob_id",
            "user": create_mock_user("1"),
        }

    @patch.object(blob_acl, "check_can_read_document")
    @patch.object(blob_acl, "Blob")
    def test_check_can_read_document_called(
        self,
        mock_blob,
        mock_check_can_read_document,
    ):
        """test_check_can_read_document_called"""
        mock_blob.get_by_id.return_value = self.mock_blob_object
        blob_acl.can_get_pid_for_blob(**self.mock_kwargs)
        mock_check_can_read_document.assert_called_with(
            self.mock_blob_object, self.mock_kwargs["user"]
        )

    @patch.object(blob_acl, "check_can_read_document")
    def test_check_can_read_document_not_called_for_superuser(
        self,
        mock_check_can_read_document,
    ):
        """test_check_can_read_document_not_called_for_superuser"""
        self.mock_kwargs["user"] = create_mock_user("1", is_superuser=True)
        blob_acl.can_get_pid_for_blob(**self.mock_kwargs)
        mock_check_can_read_document.assert_not_called()

    @patch.object(blob_acl, "check_can_read_document")
    @patch.object(blob_acl, "Blob")
    def test_check_can_read_document_exception_fails(
        self,
        mock_blob,
        mock_check_can_read_document,
    ):
        """test_check_can_read_document_exception_fails"""
        mock_blob.get_by_id.return_value = self.mock_blob_object
        mock_check_can_read_document.side_effect = AccessControlError(
            "mock_check_can_read_document_acl_error"
        )

        with self.assertRaises(AccessControlError):
            blob_acl.can_get_pid_for_blob(**self.mock_kwargs)

    @patch.object(blob_acl, "check_can_read_document")
    @patch.object(blob_acl, "Blob")
    def test_success_returns_function_output(
        self,
        mock_blob,
        mock_check_can_read_document,  # noqa, pylint: disable=unused-argument
    ):
        """test_success_returns_function_output"""
        mock_blob.get_by_id.return_value = self.mock_blob_object

        self.assertEqual(
            blob_acl.can_get_pid_for_blob(**self.mock_kwargs),
            self.mock_func_retval,
        )
        self.mock_kwargs["func"].assert_called_with(
            self.mock_kwargs["blob_id"], self.mock_kwargs["user"]
        )


class TestCanSetPidForBlob(TestCase):
    """Unit tests for `can_set_pid_for_blob` function."""

    def setUp(self) -> None:
        """setUp"""
        self.mock_blob_object = "mock_blob_object"
        self.mock_func_retval = "mock_func_retval"

        mock_func = MagicMock()
        mock_func.return_value = self.mock_func_retval

        self.mock_kwargs = {
            "func": mock_func,
            "blob_id": "mock_blob_id",
            "blob_pid": "mock_blob_pid",
            "user": create_mock_user("1"),
        }

    @patch.object(blob_acl, "check_can_read_document")
    @patch.object(blob_acl, "Blob")
    def test_check_can_read_document_called(
        self,
        mock_blob,
        mock_check_can_read_document,
    ):
        """test_check_can_read_document_called"""
        mock_blob.get_by_id.return_value = self.mock_blob_object
        blob_acl.can_set_pid_for_blob(**self.mock_kwargs)
        mock_check_can_read_document.assert_called_with(
            self.mock_blob_object, self.mock_kwargs["user"]
        )

    @patch.object(blob_acl, "check_can_read_document")
    def test_check_can_read_document_not_called_for_superuser(
        self,
        mock_check_can_read_document,
    ):
        """test_check_can_read_document_not_called_for_superuser"""
        self.mock_kwargs["user"] = create_mock_user("1", is_superuser=True)
        blob_acl.can_set_pid_for_blob(**self.mock_kwargs)
        mock_check_can_read_document.assert_not_called()

    @patch.object(blob_acl, "check_can_read_document")
    @patch.object(blob_acl, "Blob")
    def test_check_can_read_document_exception_fails(
        self,
        mock_blob,
        mock_check_can_read_document,
    ):
        """test_check_can_read_document_exception_fails"""
        mock_blob.get_by_id.return_value = self.mock_blob_object
        mock_check_can_read_document.side_effect = AccessControlError(
            "mock_check_can_read_document_acl_error"
        )

        with self.assertRaises(AccessControlError):
            blob_acl.can_set_pid_for_blob(**self.mock_kwargs)

    @patch.object(blob_acl, "check_can_read_document")
    @patch.object(blob_acl, "Blob")
    def test_success_returns_function_output(
        self,
        mock_blob,
        mock_check_can_read_document,  # noqa, pylint: disable=unused-argument
    ):
        """test_success_returns_function_output"""
        mock_blob.get_by_id.return_value = self.mock_blob_object

        self.assertEqual(
            blob_acl.can_set_pid_for_blob(**self.mock_kwargs),
            self.mock_func_retval,
        )
        self.mock_kwargs["func"].assert_called_with(
            self.mock_kwargs["blob_id"],
            self.mock_kwargs["blob_pid"],
            self.mock_kwargs["user"],
        )



""" System API to manage  Blob objects.
"""
import logging

from core_linked_records_app import settings
from core_linked_records_app.components.local_id.models import LocalId
from core_linked_records_app.system.local_id import api as local_id_system_api
from core_linked_records_app.utils.path import get_api_path_from_object
from core_linked_records_app.utils.providers import delete_record_from_provider
from core_main_app.commons import exceptions
from core_main_app.commons.exceptions import DoesNotExist
from core_main_app.components.blob.models import Blob
from core_main_app.system.blob import api as blob_system_api

logger = logging.getLogger(__name__)


def get_pid_for_blob(blob_id):
    """Retrieve PID matching the blob ID provided.

    Args:
        blob_id:

    Raises:
        DoesNotExist: No PID found for the given Blob ID

    Returns:
        str - PID of the blob object
    """
    try:
        return local_id_system_api.get_by_class_and_id(
            record_object_class=get_api_path_from_object(Blob()),
            record_object_id=blob_id,
        )
    except exceptions.DoesNotExist as dne:
        raise exceptions.DoesNotExist(str(dne))
    except Exception as exc:
        error_message = f"An error occurred while looking up PID assigned to blob '{blob_id}'"

        logger.error("%s: %s", error_message, str(exc))
        raise exceptions.ApiError(error_message)


def set_pid_for_blob(blob_id, blob_pid):
    """Retrieve PID matching the blob ID provided.

    Args:
        blob_id:
        blob_pid:
    """
    try:
        record_name = (
            f"{settings.ID_PROVIDER_PREFIX_BLOB}/{blob_pid.split('/')[-1]}"
        )

        try:
            local_id_object = get_pid_for_blob(blob_id)
        except exceptions.DoesNotExist:
            try:
                local_id_object = local_id_system_api.get_by_name(record_name)
            except exceptions.DoesNotExist:
                local_id_object = None

        if local_id_object:
            local_id_object.record_name = record_name
            local_id_object.record_object_class = get_api_path_from_object(
                Blob()
            )
            local_id_object.record_object_id = str(blob_id)
        else:
            local_id_object = LocalId(
                record_name=record_name,
                record_object_class=get_api_path_from_object(Blob()),
                record_object_id=str(blob_id),
            )

        return local_id_system_api.insert(local_id_object)
    except Exception as exc:
        error_message = f"An error occurred while assigning PID '{blob_pid}' to blob '{blob_id}'"

        logger.error("%s: %s", error_message, str(exc))
        raise exceptions.ApiError(error_message)


def delete_pid_for_blob(blob: Blob):
    """Deletes the PID assigned to the blob passed in parameter. If no PID has
    been assigned, the function simply exits.

    Args:
        blob (Blob): The blob for which the PID needs to be deleted.
    """
    try:
        local_id_obj: LocalId = local_id_system_api.get_by_class_and_id(
            get_api_path_from_object(Blob()), blob.pk
        )

        # Delete the PID using the internal name.
        delete_record_from_provider(local_id_obj.record_name)
    except DoesNotExist:  # If there is no previous PID assigned.
        logger.info(
            "No PID assigned to the blob %s (%s). Skipping deletion.",
            str(blob.pk),
            blob.filename,
        )
        return


def get_blob_by_pid(pid):
    """Return blob object with the given pid.

    Args:
        pid (str): PID of the blob.

    Raises:
        DoesNotExist: The PID is not assigned to any Blob object.
        AccessControlError: The Blob cannot be accessed by the user.
        ApiError: Any other error occured while trying to resolve the PID.

    Returns:
        Blob: Blob object assigned to the given PID.
    """
    try:
        # From the PID url (e.g. https://pid-system.org/prefix/record), retrieve
        # only the prefix and record (e.g. prefix/record) stored in DB.
        pid_internal_name = "/".join(pid.split("/")[-2:])
        local_id_object = local_id_system_api.get_by_name(pid_internal_name)

        # Ensure the LocalID object refers to a Blob
        if not (
            local_id_object.record_object_class
            and local_id_object.record_object_id
        ):
            raise exceptions.DoesNotExist("PID not assigned to blob")

        # Retrieve the blob object using the blob system API.
        # FIXME should be made more generic (using local_id_object.record_object_class)
        #  to retrieve any kind of object.
        return blob_system_api.get_by_id(local_id_object.record_object_id)
    except exceptions.DoesNotExist as not_assigned_exc:
        error_message = f"PID '{pid}' not assigned to blob"
        logger.error(error_message)
        raise exceptions.DoesNotExist(error_message) from not_assigned_exc
    except Exception as exc:
        error_message = (
            f"An error occurred while looking up blob assigned to PID '{pid}'"
        )

        logger.error("%s: %s", error_message, str(exc))
        raise exceptions.ApiError(error_message) from exc



""" Custom admin site for the LocalId model
"""
from django.contrib import admin


class CustomLocalIdAdmin(admin.ModelAdmin):
    """CustomLocalIdAdmin"""

    def has_add_permission(
        self, request, obj=None
    ):  # pylint: disable=unused-argument
        """Prevent from manually adding LocalID objects"""
        return False

    def has_change_permission(self, request, obj=None):
        """Prevent from manually editing LocalID objects"""
        return False



""" Signals to trigger after Blob modificaations.
"""
import json
from logging import getLogger

from django.db import transaction
from django.db.models.signals import post_save, post_delete
from django.urls import resolve, Resolver404
from rest_framework import status

from core_linked_records_app import settings
from core_linked_records_app.components.pid_settings.models import PidSettings
from core_linked_records_app.system.blob import api as blob_system_api
from core_linked_records_app.utils import exceptions
from core_linked_records_app.utils.pid import split_prefix_from_record
from core_linked_records_app.utils.providers import ProviderManager
from core_main_app.commons.exceptions import CoreError, DoesNotExist
from core_main_app.components.blob.models import Blob

logger = getLogger(__name__)


def init():
    """Connect to Blob object events."""
    post_save.connect(set_blob_pid, sender=Blob)
    post_delete.connect(delete_blob_pid, sender=Blob)


def _register_pid_for_blob_id(provider_name, pid_value, blob_id):
    """Registers a blob with a set PID to a given provider.

    Args:
        provider_name:
        pid_value:
        blob_id:

    Returns:
        str - Persistent identifier
    """
    # Retrieve provider, prefix and record information to register the blob in
    # the default provider.
    try:
        # Retrieve the current provider.
        provider_manager = ProviderManager()
        provider = provider_manager.get(provider_name)
    except Exception as exc:
        error_message = (
            "Error while retrieving the provider for registring the PID"
        )
        logger.error("%s: %s", error_message, str(exc))
        raise exceptions.InvalidProviderError(error_message)

    # Parse the `pid_value` to obtain record & prefix info.
    local_pid_url = pid_value.replace(
        provider.provider_lookup_url, provider.local_url
    )

    # Try resolving the `local_pid_url` and ensure the expected view was
    # retrieved.
    try:
        resolver_match = resolve(
            local_pid_url.replace(settings.SERVER_URI, "")
        )

        if resolver_match.view_name != "core_linked_records_provider_record":
            raise Resolver404()
    except Resolver404 as exc:
        error_message = "Invalid view retrieved while registring the PID"
        logger.error(error_message)
        raise exceptions.PidResolverError(error_message) from exc

    # Before asking the provider to create a record, separate prefix from
    # record name.
    try:
        prefix, record = split_prefix_from_record(
            resolver_match.kwargs["record"]
        )
    except (
        exceptions.InvalidPrefixError,
        exceptions.InvalidRecordError,
    ) as exc:
        error_message = "Error while retrieving prefix from the PID record"
        logger.error("%s: %s", error_message, str(exc))
        raise exceptions.InvalidPidError(error_message)

    provider_response = provider.create(prefix, record)

    # If an error happened during PID registration, try to relay the message
    # from the provider, otherwise relay a default error message.
    default_error_message = "An error occurred while creating the PID"
    try:
        if (
            provider_response.status_code
            not in [status.HTTP_200_OK, status.HTTP_201_CREATED]
            and blob_system_api.get_blob_by_pid(pid_value).pk != blob_id
        ):
            error_message = json.loads(provider_response.content).get(
                "message", default_error_message
            )
            logger.error(error_message)
            raise exceptions.PidCreateError(error_message)
    except Exception as exc:  # If the response is not JSON parsable
        logger.error("%s: %s", default_error_message, str(exc))
        raise exceptions.PidCreateError(default_error_message)

    return json.loads(provider_response.content)["url"]


def _set_blob_pid(instance: Blob):
    """Set the PID in the given Blob `instance`. If the PID
    already exists and is valid, it is not reset.

    Args:
        instance:

    Raises:
        CoreError: If any exception occur while executing the function.
    """
    try:
        if not PidSettings.get().auto_set_pid:
            return

        try:
            blob_system_api.get_pid_for_blob(str(instance.pk))
        except DoesNotExist:
            # Create default PID value.
            default_pid_value = (
                f"{ProviderManager().get(settings.ID_PROVIDER_SYSTEM_NAME).provider_lookup_url}/"
                f"{settings.ID_PROVIDER_PREFIX_DEFAULT}"
            )

            # Register PID and write resulting URL in instance.
            pid_value = _register_pid_for_blob_id(
                settings.ID_PROVIDER_SYSTEM_NAME,
                default_pid_value,
                instance.pk,
            )
            blob_system_api.set_pid_for_blob(instance.pk, pid_value)
    except Exception as exc:
        error_message = (
            f"An error occurred while setting a PID for blob '{instance.pk}'"
        )

        logger.error("%s: %s", error_message, str(exc))
        raise CoreError(f"{error_message}.") from exc


def set_blob_pid(
    sender, instance: Blob, **kwargs  # noqa, pylint: disable=unused-argument
):
    """Wrapper around `_set_blob_pid` to ensure the PID can be updated while
    the Blob is being saved without locking the DB.

    Args:
        sender:
        instance:
        kwargs:
    """
    transaction.on_commit(lambda: _set_blob_pid(instance))


def delete_blob_pid(
    sender, instance: Blob, **kwargs  # noqa, pylint: disable=unused-argument
):
    """Delete a PID assigned to a Blob

    Args:
        sender:
        instance:
        kwargs:
    """
    try:
        blob_system_api.delete_pid_for_blob(instance)
    except Exception as exc:  # pylint: disable=broad-except
        logger.warning(
            "Trying to delete PID for blob %d (%s) but an error occurred: %s",
            instance.pk,
            instance.filename,
            str(exc),
        )



""" REST views for the blob API
"""
from rest_framework import status
from rest_framework.permissions import (
    IsAdminUser,
)
from rest_framework.response import Response

from core_linked_records_app.components.blob import api as blob_api
from core_linked_records_app.system.local_id import api as local_id_system_api
from core_main_app.commons.exceptions import CoreError, DoesNotExist
from core_main_app.rest.blob.views import BlobList


class BlobUploadWithPIDView(BlobList):
    """Blob Upload With PID View"""

    permission_classes = (IsAdminUser,)

    def post(self, request):
        """Upload a Blob with a PID. Available only for superusers.

        Args:
            request: HTTP request

        Returns:
            - code: 200
              content: Created blob
            - code: 400
              content: Validation error
            - code: 500
              content: Internal server error
        """
        if not request.user.is_superuser:
            return Response(
                {"message": "Only a superuser can use this feature."},
                status=status.HTTP_403_FORBIDDEN,
            )

        try:
            pid = request.POST.get("pid", None)

            if pid is None:
                raise CoreError("Missing PID field in POST data.")

            # Check that the PID has not yet been assigned.
            try:
                # Rebuild the PID 'prefix/record' by extracting the last 2 items
                # of the PID path and performs the lookup.
                local_id_system_api.get_by_name("/".join(pid.split("/")[-2:]))
                raise CoreError("PID has already been assigned.")
            except DoesNotExist:
                pass

            # Upload the blob and return the error if there is one.
            blob_upload_response = super().post(request)

            if blob_upload_response.status_code != status.HTTP_201_CREATED:
                return blob_upload_response

            serialized_data = blob_upload_response.data
            serialized_data["pid"] = pid

            # Assign PID to blob
            blob_api.set_pid_for_blob(
                serialized_data["id"], serialized_data["pid"], request.user
            )

            # Return the serialized data
            return Response(serialized_data, status=status.HTTP_201_CREATED)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(
                content, status=status.HTTP_500_INTERNAL_SERVER_ERROR
            )



""" Linked records PID Path objects.
"""
import logging

from django.core.exceptions import ObjectDoesNotExist
from django.db import models

from core_linked_records_app.utils.dict import validate_dot_notation
from core_main_app.commons import exceptions
from core_main_app.commons.exceptions import ModelError
from core_main_app.components.template.models import Template

logger = logging.getLogger(__name__)


class PidPath(models.Model):
    """PidPath model"""

    path = models.CharField(
        blank=False, max_length=255, validators=[validate_dot_notation]
    )
    template = models.OneToOneField(
        Template,
        blank=False,
        on_delete=models.CASCADE,
        null=False,
        unique=True,
    )

    @staticmethod
    def get_all():
        """Retrieve all PidPath objects.

        Returns:
        """
        try:
            return PidPath.objects.all()  # pylint: disable=no-member
        except Exception as exc:
            raise ModelError(str(exc)) from exc

    @staticmethod
    def get_all_by_template_list(template_list):
        """Retrieve a list of PidPath given a list of templates.

        Args:
            template_list:

        Returns:
        """
        try:
            return PidPath.objects.filter(  # pylint: disable=no-member
                template__in=template_list
            )
        except Exception as exc:
            raise ModelError(str(exc)) from exc

    @staticmethod
    def get_by_template(template):
        """Return all PidPath defined for a given template.

        Args:
            template:

        Returns:
        """
        try:
            return PidPath.objects.get(  # pylint: disable=no-member
                template=template
            )
        except ObjectDoesNotExist:
            return None
        except Exception as exc:
            raise exceptions.ModelError(str(exc)) from exc

    def __str__(self):
        """PidPath object as string.

        Returns:
            str - String representation of PidPath object.
        """
        return f"PID path '{self.path}' for template '{self.template}'"
